using System;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;
using System.Xml;
using System.Xml.Linq;


namespace EnhancerEvaluation
{
    public class Config : ConfigBase
    {
        public static Dictionary<string, string> DataSources = new Dictionary<string, string>
        {
            { "Histone", "HistoneListFileName" },
            { "AllChIP", "AllChIPListFileName" },
            { "TRF", "TRFListFileName" },
        };

        public Config(string xmlFile) : base(xmlFile)
        {

        }

        public string GetChipDataSourceFileList(string dataSource)
        {
            return Document.Descendants(DataSources[dataSource]).First().Value.ToString();
        }

        public string GetChipDataSourceFileList(string dataSource, string cellLine)
        {
            return this.CellLineTag.Replace(Document.Descendants(DataSources[dataSource]).First().Value.ToString(), cellLine);
        }

        public string TssFileName
        {
            get { return Document.Descendants("TssFileName").First().Value.ToString(); }
        }

        public string DrmFileName
        {
            get { return Document.Descendants("DrmFileName").First().Value.ToString(); }
        }

        public string GetDrmFileNameForCellLine(string cellLine)
        {
            return this.CellLineTag.Replace(DrmFileName, cellLine);
        }

        public string MapFileFormat
        {
            get
            {
                var mapData = Document.Descendants("MapFileFormat").FirstOrDefault();
                return mapData == null ? null : mapData.Value.ToString();
            }
        }

        public string ExpressionFileFormat
        {
            get { return Document.Descendants("ExpressionFileFormat").First().Value.ToString(); }
        }

        public string GetExpressionFileName(string rnaSource, string cellLine)
        {
            return this.RnaSourceTag.Replace(this.CellLineTag.Replace(ExpressionFileFormat, cellLine), rnaSource);
        }

        public string OutFileFormat
        {
            get { return Document.Descendants("OutputFileFormat").First().Value.ToString(); }
        }

        public string GetGenomeWideExpressionFile(string cellLine)
        {
            return string.Format(Document.Descendants("GenomeWideRegressionFile").First().Value, cellLine);
        }

        public double D0
        {
            get { return double.Parse(Document.Descendants("D0").First().Value.ToString()); }
        }

        public int Range
        {
            get { return int.Parse(Document.Descendants("PromoterRange").First().Value.ToString()); }
        }

        public int SourceCount
        {
            get { return Document.Descendants("RnaSources").First().Value.ToString().Split(',').Count(); }
        }
            
        public string[] CellLines
        {
            get
            {
                if (cellLines == null )
                {
                    var cellLineData = Document.Descendants("CellLines").FirstOrDefault();
                    cellLines = cellLineData == null ? null : cellLineData.Value.Split(',').ToArray();
                }
                return cellLines;
            }
        }
         
        public string[] RnaSources
        {
            get
            {
                if (rnaSources == null)
                {
                    var rnaSourceData = Document.Descendants("RnaSources").FirstOrDefault();
                    rnaSources = rnaSourceData == null ? null : rnaSourceData.Value.Split(',').ToArray();
                }
                return rnaSources;
            }
        }

        public string[] Thresholds
        {
            get
            {
                if (thresholds == null)
                {
                    var thresholdData = Document.Descendants("SourceThreshholds").FirstOrDefault();
                    thresholds = thresholdData == null ? null : thresholdData.Value.Split(',').ToArray();
                }
                return thresholds;
            }
        }

        public Regex CellLineTag
        {
            get
            { 
                if (cellLineTag == null)
                {
                    cellLineTag = new Regex(Document.Descendants("CellLineTag").First().Value); 
                }
                return cellLineTag;
            }
        }

        public Regex ThresholdTag
        {
            get
            { 
                if (thresholdTag == null)
                {
                    thresholdTag = new Regex(Document.Descendants("ThreshholdTag").First().Value); 
                }
                return thresholdTag;
            }
        }

        public Regex RnaSourceTag
        {
            get
            { 
                if (rnaSourceTag == null)
                {
                    rnaSourceTag = new Regex(Document.Descendants("RnaTag").First().Value); 
                }
                return rnaSourceTag;
            }
        }

        public string[][] GetMapFileNames(string cellLine)
        {
            return RnaSources.Select(rnaSource => 
                       Thresholds.Select(threshhold =>
                          CellLineTag.Replace(
                              RnaSourceTag.Replace(
                                  ThresholdTag.Replace(MapFileFormat, threshhold), 
                                  rnaSource),  
                              cellLine))
                           .ToArray())
                       .ToArray();
        }

        public string[][] GetTableFileNames(string cellLine)
        {
            return RnaSources.Select(source =>
                       Thresholds.Select(threshold => 
                           cellLineTag.Replace(ThresholdTag.Replace(RnaSourceTag.Replace(
                               OutFileFormat, 
                               source),
                           threshold), 
                       cellLine)).ToArray()).ToArray();
        }

        //private XDocument Document;

        private string[] cellLines;
        private string[] rnaSources;
        private string[] thresholds;

        private Regex cellLineTag;
        private Regex thresholdTag;
        private Regex rnaSourceTag;
    }
}

