using System;
using System.Linq;
using System.Xml;
using System.Xml.Linq;
using System.Collections.Generic;
using System.IO;
using System.Text.RegularExpressions;

namespace EnhancerEvaluation
{
	class MainClass
	{
		public static void Main (string[] args)
		{
            Args a = new Args(args);

            switch (a.StringArgs["Mode"])
            {
                case "MapNearestGeneContent":   
                    MapContent(a.StringArgs["MapFileName"], a.StringArgs["GeneFileName"], a.StringArgs["OutputFileName"], false);
                    break;

                case "MapNearestDRMContent":
                    MapContent(a.StringArgs["MapFileName"], a.StringArgs["GeneFileName"], a.StringArgs["OutputFileName"], true);
                    break;
                
                case "ScoreDrmPeaks":
                    ScoreDrmPeaks(a.StringArgs["DrmFileName"], a.StringArgs["PeakFileName"], a.StringArgs["OutputFileName"]);
                    break;

                case "ScoreTssPeaks":
                    ScoreTssPeaks(
                        a.StringArgs["TssFileName"], 
                        a.StringArgs["PeakFileName"], 
                        a.StringArgs["OutputFileName"],
                        (double)a.IntArgs["D0"],
                        a.IntArgs["Range"]);
                    break;

                case "ScoreTssDrmNNPeaks":
                    ScoreTssDrmNNPeaks(
                        a.StringArgs["TssFileName"], 
                        a.StringArgs["DrmFileName"], 
                        a.StringArgs["PeakFileName"], 
                        a.StringArgs["ExpFileName"],
                        a.StringArgs["OutputFileName"],
                        (double)a.IntArgs["D0"],
                        a.IntArgs["Range"]);
                    break;

                case "Regression":
                    RegressionDataUseXmlConfig(a.StringArgs["Config"]);
                    break;

                case "MultipleRegression":
                    int batch;
                    bool fHasBatch = a.IntArgs.TryOptional("Batch", out batch);
                    batch = fHasBatch ? batch : 1;

                    RegressionStage stage;
                    if (!Enum.TryParse(a.StringArgs["Stage"], out stage))
                    {
                        Console.WriteLine("Bad stage value given: " + a.StringArgs["Stage"]);
                        Console.WriteLine("\tStage options are: " + string.Join(", ", Enum.GetNames(typeof(RegressionStage))));
                        return;
                    }

                    MultipleRegressionDataUseXmlConfig(
                        a.StringArgs ["Config"], 
                        a.StringArgs["DataSource"],
                        a.StringArgs.GetOptional("CellLine"),
                        a.StringArgs.GetOptional("GenerateFeatures") == "True",
                        a.StringArgs.GetOptional("SampleDRMs") == "True",
                        a.StringArgs.GetOptional("UseFullMap") == "True",
                        a.IntArgs.GetOptional("StartRep"),
                        a.IntArgs.GetOptional("RepCount"),
                        batch,
                        stage);
                    break;

                case "BuildMap":
                    Console.WriteLine(a.StringArgs["CellLine"]);
                    BuildMap(
                        a.StringArgs["Config"],
                        a.StringArgs["CellLine"],
                        a.StringArgs.GetOptional("BedFile"),
                        a.StringArgs.GetOptional("OmitDrmSource") == "True");
                    break;

                case "GenomeWideRegression":
                    GenomeWideRegression(
                        a.StringArgs["Config"], 
                        a.StringArgs["CellLine"], 
                        a.StringArgs.GetOptional("GenerateFeatures") == "True",
                        a.StringArgs.GetOptional("UseVariantExpressedSubset") == "True");
                    break;

                case "FilterTranscripts":
                    Console.WriteLine(a.StringArgs.GetOptional("OmittedCellLine"));
                    FilterTranscripts(a.StringArgs["Config"], a.StringArgs.GetOptional("OmittedCellLine"));
                    break;
                
                case "CorrelateExpression":
                    CrossCellLineExpressionCorrelation(a.StringArgs["Config"]);
                    break;

                case "TFGenePairs":
                    TFGenePairs(
                        a.StringArgs["Config"],
                        a.StringArgs["CellLine"],
                        a.StringArgs.GetOptional("GeneSets") == "True",
                        a.StringArgs.GetOptional("GetCorrelation") == "True");
                    break;

                case "MapNNOverlap":
                    MapNNOverlap(a.StringArgs["Config"], a.StringArgs["CellLine"]);
                    break;

                case "HistoneState":
                    HistoneState(a.StringArgs["Config"], a.StringArgs["CellLine"], a.StringArgs.GetOptional("ExtractData") == "True");
                    break;

                case "HistoneProfile":
                    HistoneProfile(a.StringArgs["Config"], a.StringArgs["CellLine"]);
                    break;

                case "ProgressivePromoterAdditions":
                    ProgressivePromoterAdditions(
                        a.StringArgs["Config"],
                        a.StringArgs["CellLine"],
                        a.StringArgs.GetOptional("UseVariantExpressedSubset") == "True");
                    break;

                default:
                    Console.WriteLine("Invalid argument: {0}", a.StringArgs["Mode"]);
                    break;
            }
		}

        /// <summary>
        /// HTML table data delimiters
        /// </summary>
        private static string th = "</th><th>";
        private static string td = "</td><td>";
        private static string trhs = "<tr><th>";
        private static string trhe = "</th></tr>";
        private static string trds = "<tr><td>";
        private static string trde = "</td></tr>";

        static void MapContent(string mapFileName, string geneFileName, string outputFileName, bool fByDrm)
        {
            BedFile drms = new BedFile(mapFileName, BedFile.Bed4Drm);
            BedFile genes = new BedFile(geneFileName, BedFile.Bed4Drm);

            var outer = genes;
            var inner = drms;

            if (fByDrm)
            {
                outer = drms;
                inner = genes;
            }

            var pairs = 
               (from chromosome in outer.Chromosomes
                 where outer.ChromosomeBinnedLocations.ContainsKey(chromosome) &&
                       inner.ChromosomeBinnedLocations.ContainsKey(chromosome)
                select 
                       (from geneBin in outer.ChromosomeBinnedLocations[chromosome]
                        select new 
                        {
                            GeneBin = geneBin.Value, 
                            Drms = inner.ChromosomeBinnedLocations[chromosome]
                                         .Where(x => Math.Abs(x.Key - geneBin.Key) <= 1)
                                         .Select(x => x.Value)  // Get the name-location pairs from the bin
                                         .SelectMany(x => x)    // Flatten the bins' pairs into one enumerable
                        })
                        .Select(x => x.GeneBin  // Convert enumerable of bins to enumerable of genes
                                       .Select(g => 
                                        new 
                                        {
                                            Gene = g, 
                                            DrmAndDistance = x.Drms.Select(d => new {Drm = d, Distance = DrmGeneDistance(d.Value, g.Value)})
                                        }))
                        .SelectMany(x => x)
                 ).SelectMany(x => x)
                 .Where(x => x.DrmAndDistance.Count() > 0)
                 .Select(x => 
                     {
                        /*int min = int.MaxValue;
                        KeyValuePair<string, BedFile.Location> minDrm;
                        foreach (var v in x.DrmAndDistance)
                        {
                            if (v.Distance < min)
                            {
                                min = v.Distance;
                                minDrm = v.Drm;
                            }
                        }*/

                        var minDrmGene = x.DrmAndDistance.OrderBy(d => d.Distance).Where((d, i) => i < 1);
                        var key = new List<KeyValuePair<string, BedFile.Location>>(){x.Gene};
                        Dictionary<string, int> distances = new Dictionary<string, int>();
                        foreach (var drm in minDrmGene)
                        {
                            string geneName = fByDrm ? drm.Drm.Key : x.Gene.Key;
                            string drmName = fByDrm ? x.Gene.Key : drm.Drm.Key;
                            string edge = string.Format("{0},{1}", geneName, drmName);
                            distances.Add(edge, drm.Distance);
                        }

                        return new
                        {
                            // Swap gene and drm if using drm as the outer source
                            Genes = fByDrm ? minDrmGene.Select(d => d.Drm).ToList() : key,
                            Drms = fByDrm ? key : minDrmGene.Select(d => d.Drm).ToList(),
                            Distances = distances
                        };
                    }
                    );

            using (TextWriter tw = new StreamWriter(outputFileName))
            {
                int n = 0;
                int count = pairs.Count();
                foreach (var x in pairs)
                {
                    string line = 
                        string.Join(
                            "\n", 
                                from gene in x.Genes
                                from drm in x.Drms
                                select 
                                    string.Format(
                                        "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}",
                                        drm.Value.Chromosome, 
                                        drm.Value.Start,
                                        drm.Value.End,
                                        drm.Value.Strand,
                                        gene.Key,
                                        drm.Key,
                                        x.Distances[string.Format("{0},{1}", gene.Key, drm.Key)]));
                    tw.WriteLine(line);
                    if (n % 10000 == 0)
                    {
                        Console.WriteLine("Processed {0} of {1} genes", n, count);
                    }
                    n++;
                }                 
            }
        }

        private static int DrmGeneDistance(BedFile.Location l1, BedFile.Location l2)
        {
            if (string.Compare(l1.Chromosome, l2.Chromosome) != 0)
            {
                return int.MaxValue;
            }

            return Math.Abs(l1.Mid - l2.Start);
        }

        static void ScoreDrmPeaks(string binFileName, string peakFileName, string outputFileName)
        {
            BedFile drmFile = new BedFile(binFileName, BedFile.Bed4Drm);

            BedFile.Layout layout = peakFileName.Contains("broad") ? BedFile.Bed6Plus3Layout : BedFile.Bed6Plus4Layout;
            BedFile peakFile = new BedFile(peakFileName, layout);

            var drmScores = TRFScorer.CreateDrmScores(
                drmFile, 
                peakFile,
                false);

            foreach (var v in drmScores.Where(x => x.Value > 0))
            {
                    Console.WriteLine("{0}\t{1}", v.Key, v.Value);
            }
        }

        static void ScoreTssPeaks(string tssFileName, string peakFileName, string outputFileName, double D0, int maxRange)
        {
            throw new NotImplementedException();
            /*BedFile tssFile = new BedFile(tssFileName, BedFile.Bed4Drm);

            BedFile.Layout layout = peakFileName.Contains("broad") ? BedFile.Bed6Plus3Layout : BedFile.Bed6Plus4Layout;
            BedFile peakFile = new BedFile(peakFileName, layout);

            var tssScores = CreateTssScores(tssFile, peakFile, D0, maxRange);
         
            foreach (var v in tssScores.Where(x => x.Value > 0))
            {
                Console.WriteLine("{0}\t{1}", v.Key, v.Value);
            }*/
        }

        /// <summary>
        /// Organize tss location data by chromosome in sequence
        /// </summary>
        /// <returns>The ordered tss starts.</returns>
        /// <param name="tssData">Tss data.</param>
        private static Dictionary<string, List<BedFile.Location>> GetOrderedTssStarts(
            Dictionary<string, BedFile.Location> tssData)
        {
            return tssData
                .GroupBy(x => x.Value.Chromosome, x => x.Value)
                .ToDictionary(x => x.Key, x => x.OrderBy(y => y.Mid).ToList());
        }

        /// <summary>
        /// Organizes tsses or drms by chromosome and position and filters to only allowed objects
        /// </summary>
        /// <returns>The ordered filtered locations.</returns>
        /// <param name="data">Data.</param>
        /// <param name="allowedItems">Allowed items.</param>
        private static Dictionary<string, List<BedFile.Location>> GetOrderedFilteredLocations(
            Dictionary<string, BedFile.Location> data,
            HashSet<string> allowedItems)
        {
            return data
                .Where(x => allowedItems.Contains(x.Value.Name))
                .ToLookup(x => x.Value.Chromosome, x => x.Value)
                .ToDictionary(x => x.Key, x => x.OrderBy(y => y.Mid).ToList());
        }

        /// <summary>
        /// Get drms that are not present in any promoters given by chromosome
        /// </summary>
        /// <returns>The drms in promoters.</returns>
        /// <param name="drmFile">Drm file.</param>
        /// <param name="orderedTssStarts">Ordered tss starts.</param>
        static Dictionary<string, List<BedFile.Location>> RemoveDrmsInPromoters(
            BedFile drmFile, 
            Dictionary<string, List<BedFile.Location>> orderedTssStarts)
        {
            return drmFile.Locations.Where(x => 
                                           {
                bool hasChromosome = orderedTssStarts.ContainsKey(x.Value.Chromosome);
                if (hasChromosome)
                {
                    BedFile.Location closestTss = TRFScorer.GetClosest(orderedTssStarts[x.Value.Chromosome], x.Value.Mid);
                    if (Math.Abs(closestTss.Start - x.Value.Start) > 10000 && Math.Abs(closestTss.Start - x.Value.End) > 10000)
                    {
                        return true;
                    }
                }
                return false; 
            }).ToLookup(x => x.Value.Chromosome, x => x.Value)
            .ToDictionary(x => x.Key, x => x.OrderBy(drm => drm.Mid).ToList());
        }

        /// <summary>
        /// Creates the tss centric NN map.
        /// </summary>
        /// <returns>The tss centric NN map.</returns>
        /// <param name="orderedTssStarts">Ordered tss starts.</param>
        /// <param name="orderedDrmStarts">Ordered drm starts.</param>
        private static IEnumerable<Tuple<string, string>> CreateTssCentricNNMap(
            Dictionary<string, List<BedFile.Location>> orderedTssStarts,
            Dictionary<string, List<BedFile.Location>> orderedDrmStarts)
        {
            return orderedTssStarts
                .Where(chr => orderedDrmStarts.ContainsKey(chr.Key))
                .SelectMany(x => x.Value)
                .Select(tss => new { Drm = TRFScorer.GetClosestExcludeRange(orderedDrmStarts[tss.Chromosome], tss.Start, 10000), Tss = tss } )
                .Where(x => Math.Abs(x.Tss.Start - x.Drm.Mid) > 10000)
                .Select(x => new Tuple<string, string>(x.Tss.Name, x.Drm.Name));
        }

        /// <summary>
        /// Creates the drm centric NN map.
        /// </summary>
        /// <returns>The drm centric NN map.</returns>
        /// <param name="orderedTssStarts">Ordered tss starts.</param>
        /// <param name="orderedDrmStarts">Ordered drm starts.</param>
        /// <param name="drmLocations">Drm file.</param>
        private static IEnumerable<Tuple<string, string>> CreateDrmCentricNNMap(
            Dictionary<string, List<BedFile.Location>> orderedTssStarts,
            Dictionary<string, List<BedFile.Location>> orderedDrmStarts,
            Dictionary<string, BedFile.Location> drmLocations)
        {
            return drmLocations
                .Where(drm => orderedTssStarts.ContainsKey(drm.Value.Chromosome))
                .Select(drm => new { Drm = drm, Tss = TRFScorer.GetClosest(orderedTssStarts [drm.Value.Chromosome], drm.Value.Mid) })
                .Where(x => Math.Abs(x.Tss.Start - x.Drm.Value.Mid) > 10000)
                .Select(x => new Tuple<string, string>(x.Tss.Name, x.Drm.Value.Name));
        }

        /// <summary>
        /// Create full bi-directional DRM and TSS centric nearest neighbor maps
        /// </summary>
        /// <returns>The drm NN map.</returns>
        /// <param name="tssData">Tss data.</param>
        /// <param name="drmFile">Drm file.</param>
        static ILookup<string, string> CreateDrmNNMap(
            Dictionary<string, BedFile.Location> tssData,
            BedFile drmFile,
            bool forwardOnly)
        {
            Dictionary<string, List<BedFile.Location>> orderedTssStarts = GetOrderedTssStarts(tssData);

            Dictionary<string, List<BedFile.Location>> orderedDrmStarts = drmFile.GetOrderedLocations(x => x.Mid);

            //IEnumerable<Tuple<string, string>> drmNNmap = CreateDrmCentricNNMap(orderedTssStarts, orderedDrmStarts, drmFile.Locations);

            //IEnumerable<Tuple<string, string>> tssNNmap = CreateTssCentricNNMap(orderedTssStarts, orderedDrmStarts);

            if (forwardOnly)
            {
                return CreateTssCentricNNMap(orderedTssStarts, RemoveDrmsInPromoters(drmFile, orderedTssStarts)).ToLookup(x => x.Item1, x => x.Item2);
            } 
            else
            {
                return CreateDrmCentricNNMap(orderedTssStarts, orderedDrmStarts, drmFile.Locations)
                    .Concat(CreateTssCentricNNMap(orderedTssStarts, orderedDrmStarts))
                    .Distinct()
                    .ToLookup(x => x.Item1, x => x.Item2);
            }
        }

        /// <summary>
        /// Create full bi-directional DRM and TSS centric nearest neighbor maps
        /// </summary>
        /// <returns>The drm NN map.</returns>
        /// <param name="tssData">Tss data.</param>
        /// <param name="drmFile">Drm file.</param>
        static ILookup<string, string> CreateDrmNNMap(
            Dictionary<string, BedFile.Location> tssData,
            BedFile drmFile,
            HashSet<string> variableTss,
            HashSet<string> variableDrm)
        {
            Dictionary<string, List<BedFile.Location>> orderedTssStarts = GetOrderedTssStarts(
                tssData.Where(pair => variableTss.Contains(pair.Key))
                          .ToDictionary(x => x.Key, x => x.Value));
            
            Dictionary<string, List<BedFile.Location>> orderedDrmStarts = drmFile.GetOrderedLocations(x => x.Mid)
                    .Select(chrLocs => new 
                    {
                        Chr = chrLocs.Key,
                        Locs = chrLocs.Value.Where(drm => variableDrm.Contains(drm.Name)).ToList()
                    })
                    .ToDictionary(x => x.Chr, x => x.Locs);

            return CreateTssCentricNNMap(orderedTssStarts, orderedDrmStarts).ToLookup(x => x.Item1, x => x.Item2);
        }

        /// <summary>
        /// Loads map filtering for TSSes that have expression data
        /// </summary>
        /// <returns>The map.</returns>
        /// <param name="mapFileName">Map file name.</param>
        /// <param name="tssData">Tss data.</param>
        static ILookup<string, string> LoadMap(string mapFileName, Dictionary<string, BedFile.Location> tssData)
        {
            using (TextReader tr = new StreamReader(mapFileName))
            {
                return tr.ReadToEnd()
                    .Split('\n')
                        .Where(line => !string.IsNullOrEmpty(line))
                        .Select(line => 
                        {
                            var fields = line.Split('\t'); 
                            return new 
                            {
                                Gene = fields[0], 
                                Drm = fields[1] 
                            };
                        })
                        .Where(x => tssData.ContainsKey(x.Gene))
                        .ToLookup(x => x.Gene, x => x.Drm);
            }
        }

        /// <summary>
        /// Loads raw map from a tsv file
        /// </summary>
        /// <returns>The map.</returns>
        /// <param name="mapFileName">Map file name.</param>
        static ILookup<string, string> LoadMap(string mapFileName)
        {
            using (TextReader tr = new StreamReader(mapFileName))
            {
                return tr.ReadToEnd()
                    .Split('\n')
                        .Where(line => !string.IsNullOrEmpty(line))
                        .Select(line => 
                                {
                            var fields = line.Split('\t'); 
                            return new 
                            {
                                Gene = fields[0], 
                                Drm = fields[1] 
                            };
                        })
                        .ToLookup(x => x.Gene, x => x.Drm);
            }
        }

        private static Dictionary<string, double> GetExpression(string expFileName)
        {
            using (TextReader tr = new StreamReader(expFileName))
            {
                return tr.ReadToEnd().Split('\n')
                        .Where(line => !string.IsNullOrWhiteSpace(line))
                        .Select(line => 
                        {
                            var f = line.Split('\t');
                            return new 
                            {
                                Gene = f [0], 
                                Exp = double.Parse(f [1])
                            };
                        })
                        .ToLookup(x => x.Gene, x => x.Exp)
                        .ToDictionary(x => x.Key, x => x.Max());
            }
        }

        private static Dictionary<string, BedFile.Location> GetExpressionTssData(string expFileName)
        {
            using (TextReader tr = new StreamReader(expFileName))
            {
                return tr.ReadToEnd().Split('\n')
                    .Where(line => !string.IsNullOrWhiteSpace(line))
                        .Select(line => 
                        {
                            string[] fields = line.Split('\t');
                            int position = int.Parse(fields[0].Split('_')[1]);
                            return new 
                            {
                                Gene = fields [0], 
                                Data = new BedFile.Location
                                {
                                    Name = fields[0],
                                    Chromosome = fields[2],
                                    Start = position,
                                    End = position,
                                }
                            };
                        })
                        .Where(x => !string.IsNullOrWhiteSpace(x.Data.Chromosome))
                        .ToLookup(x => x.Gene, x => x.Data)
                        .ToDictionary(x => x.Key, x => x.First());
            }
        }

        static void ScoreTssDrmNNPeaks(
            string tssFileName,
            string drmFileName, 
            string peakFileName,
            string expFileName,
            string outputFileName,
            double D0,
            int range)
        {
            BedFile tssFile = new BedFile(tssFileName, BedFile.Bed4Drm);
            BedFile drmFile = new BedFile(drmFileName, BedFile.Bed4Drm);

            BedFile.Layout layout = peakFileName.Contains("broad") ? BedFile.Bed6Plus3Layout : BedFile.Bed6Plus4Layout;
            BedFile peakFile = new BedFile(peakFileName, layout);

            var drmMap = CreateDrmNNMap(tssFile.Locations, drmFile, false);

            Console.WriteLine("Map created\n");

            var drmScores = TRFScorer.CreateDrmScores(drmFile, peakFile, false);

            var tssScores = TRFScorer.CreateTssScores(tssFile.Locations, peakFile, D0, range);

            Dictionary<string, double> expression = GetExpression(expFileName);

            var v = from gene in expression
                join tss in tssScores on gene.Key equals tss.Key
                select new
            {
                Tss = gene.Key,
                TssOnlyScore = tss.Value,
                Score = (drmMap.Contains(tss.Key) ? drmMap [tss.Key].Sum(x => drmScores [x]) : 0.0),
                Exp = gene.Value
            };


            using (TextWriter tw = new StreamWriter(outputFileName))
            { 
                tw.WriteLine("\"\",\"Gene\",\"TssScore\",\"TssDrmScore\",\"Expression\"");
                int i = 0;
                foreach (var tssScore in v.Where(x => drmMap.Contains(x.Tss)))
                {
                    tw.WriteLine("\"{0}\",{1},{2},{3},{4}", i, tssScore.Tss, tssScore.TssOnlyScore, tssScore.Score, tssScore.Exp);
                    i++;
                }
            }
        }

        private static ILookup<string, int> GetTSSMap(IEnumerable<string> geneIDs)
        {
            return geneIDs.Select(x => 
            {
                var fields = x.Split('_');
                return new
                {
                    GeneName = fields[0],
                    TSSLocation = int.Parse(fields[1])
                };
            }).ToLookup(x => x.GeneName, x => x.TSSLocation);
        }

        private static Dictionary<string, string> NearestTSS(IEnumerable<string> geneIDs, ILookup<string, int> TSSlocs)
        {
            return geneIDs.Select(x => 
            {
                var fields = x.Split('_');
                string GeneId = fields[0];
                int location = int.Parse(fields[1]);
                
                int minDistance = int.MaxValue;
                int tssStart = int.MaxValue;
                if (TSSlocs.Contains(GeneId))
                {
                    foreach (int tssLocation in TSSlocs[GeneId])
                    {
                        if (Math.Abs(location - tssLocation) < minDistance)
                        {
                            minDistance = Math.Abs(location - tssLocation);
                            tssStart = tssLocation;
                        }
                    }
                }

                return new
                {
                    GeneId = x,
                    TSSGeneId = string.Format("{0}_{1}", GeneId, tssStart),
                    TssStart = tssStart
                };
            }).Where(x => x.TssStart != int.MaxValue && x.GeneId == x.TSSGeneId)
            .ToDictionary(x => x.GeneId, x => x.TSSGeneId);
        }

        private static void RegressionDataUseXmlConfig(string xmlFile)
        {
            XDocument doc = XDocument.Load(xmlFile);

            string tssFileName = doc.Descendants("TssFileName").First().Value.ToString();
            string chipFileName = doc.Descendants("ChIPFileListFileName").First().Value.ToString();
            string drmFileName = doc.Descendants("DrmFileName").First().Value.ToString();
            var mapData = doc.Descendants("MapFileName").FirstOrDefault();
            string mapFileName = mapData == null ? null : mapData.Value.ToString();
            string expFileName = doc.Descendants("ExpressionFileName").First().Value.ToString();
            string outFileName = doc.Descendants("OutputFileName").First().Value.ToString();
            double D0 = double.Parse(doc.Descendants("D0").First().Value.ToString());
            int range = int.Parse(doc.Descendants("PromoterRange").First().Value.ToString());

            GetRegressionData(
                tssFileName,
                chipFileName,
                drmFileName,
                mapFileName,
                expFileName,
                outFileName,
                D0,
                range);
        }

        private static void GetRegressionData(
            string tssFileName,
            string chipFileName,
            string drmFileName,
            string mapFileName,
            string expFileName,
            string outFileName,
            double D0,
            int range)
        {
            BedFile tssFile = new BedFile(tssFileName, BedFile.Bed4Drm);
            BedFile drmFile = new BedFile(drmFileName, BedFile.Bed4Drm);

            Dictionary<string, List<string>> drmMap;
            Dictionary<string, List<string>> drmMapNN;

            LoadMaps(mapFileName, tssFile.Locations, drmFile, CreateDrmNNMap(tssFile.Locations, drmFile, false), out drmMap, out drmMapNN);

            List<Dictionary<string, double>> drmScores = new List<Dictionary<string, double>>();
            List<Dictionary<string, double>> tssScores = new List<Dictionary<string, double>>();
            HashSet<string> geneIds = new HashSet<string>();

            List<string> chipDataSets = new List<string>();

            foreach (var chipFile in LoadChIPData(chipFileName))
            {
                drmScores.Add(TRFScorer.CreateDrmScores(drmFile, chipFile, false));
            
                var t = TRFScorer.CreateTssScores(tssFile.Locations, chipFile, D0, range);
                tssScores.Add(t);

                geneIds.UnionWith(tssScores.Last().Keys);
                chipDataSets.Add(chipFile.FileName);
            }

            Dictionary<string, double> expression = GetExpression(expFileName);

            ILookup<string, int> ExpressionTSSLocations = GetTSSMap(expression.Keys);

            Dictionary<string, string> ChIPToExpressionTSS = NearestTSS(drmMap.Select(x => x.Key), ExpressionTSSLocations);

            R.CreateTable(drmMap, tssScores, drmScores, expression, ChIPToExpressionTSS, outFileName, chipDataSets);

            Console.WriteLine(string.Join("\t", R.RegressMapScoreData(outFileName, chipDataSets, R.RegressionFunction.PrincipalComponents).Select(x => string.Format("{0}", x))));

            R.CreateTable(drmMapNN, tssScores, drmScores, expression, ChIPToExpressionTSS, outFileName.Replace(".tsv", "NN.tsv"), chipDataSets);
            
            Console.WriteLine(string.Join("\t", R.RegressMapScoreData(outFileName.Replace(".tsv", "NN.tsv"), chipDataSets, R.RegressionFunction.PrincipalComponents).Select(x => string.Format("{0}", x))));
        }

        static IEnumerable<BedFile> LoadChIPData(string chipFileName)
        {
            using (TextReader tr = new StreamReader(chipFileName))
            {
                string contents = tr.ReadToEnd();

                var d = contents
                    .Split('\n')
                        .Where(line => !string.IsNullOrWhiteSpace(line))
                        .Select(line => 
                {
                    string fileName = line.Trim().Split('/').Last().Split('.').First();
                    Console.WriteLine(fileName);
                    return new 
                            {
                                FileName = fileName,
                                File = new BedFile(
                                    line.Trim(), 
                                    line.Trim().Contains("broad") ? BedFile.Bed6Plus3Layout : BedFile.Bed6Plus4Layout)
                            };
                });
                return d.Select(x => x.File);
            }
        }

        static IEnumerable<string> GetChIPDataFileNames(string chipFileName)
        {  
            using (TextReader tr = new StreamReader(chipFileName))
            {
                return tr.ReadToEnd()
                    .Split('\n')
                        .Where(line => !string.IsNullOrWhiteSpace(line))
                        .Select(line => line.Trim());
            }
        }

        static IEnumerable<string> GetChIPDataSetNames(string chipFileName)
        {  
            return GetChIPDataFileNames(chipFileName).Select(x => x.Split('/').Last().Split('.').First());
        }

        static void LoadMaps(
            string mapFileName, 
            Dictionary<string, BedFile.Location> tssData,
            BedFile drmFile,
            ILookup<string, string> drmMapNNLookup,
            out Dictionary<string, List<string>> definedMap, 
            out Dictionary<string, List<string>> nearestSubMap)
        {
            ILookup<string, string> drmMapLookup = LoadMap(mapFileName, tssData);
            
            // Select the nearest neighbor map for only the genes in the Yip map
            nearestSubMap = drmMapLookup.Select(x => 
            {
                List<string> drms = drmMapNNLookup.Contains(x.Key) ? new List<string>(drmMapNNLookup[x.Key]) : new List<string>();
                return new 
                {
                    Gene = x.Key,
                    Drms = drms
                };
            }).ToDictionary(x => x.Gene, x => x.Drms);
            
            definedMap = drmMapLookup
                .Select(x => new { Key = x.Key, Value = new List<string>(x) })
                    .ToDictionary(x => x.Key, x => x.Value);
        }

        private static List<Dictionary<string, BedFile.Location>> GetTssDataFromExpression(
            string[] rnaSources,
            Func<int, string> CreateFileName)
        {
            return rnaSources.Select((x, i) => 
                {
                    string file = CreateFileName(i);
                    Console.WriteLine("\tProcessing file {0}", file);
                    return GetExpressionTssData(file);
                }).ToList();
        }

        /// <summary>
        /// Get variantly expressed TSSes or variantly histone marked DRMs
        /// </summary>
        /// <returns>The varied element.</returns>
        /// <param name="element">Element.</param>
        /// <param name="cellLine">Cell line.</param>
        /// <param name="rnaSource">Rna source.</param>
        /// <param name="useFullMap">Leave one out or all tissue sources</param>
        private static HashSet<string> GetVariedElement(string element, string cellLine, string rnaSource, bool useFullMap)
        {
            using (TextReader tr = new StreamReader("Maps/" + (useFullMap ? "Full" : "InHouse") + "/Correlation/" + cellLine + ".All." + element + "." + rnaSource + ".nsv"))
            {
                return new HashSet<string>(tr.ReadToEnd().Split('\n'));
            }
        }

        /// <summary>
        /// Get tsses with varied expression from all rna sources
        /// </summary>
        /// <returns>The tss with varied expression.</returns>
        /// <param name="cellLine">Cell line.</param>
        /// <param name="RnaSources">Rna sources.</param>
        /// <param name="useFullMap">Leave one out or all tissue sources</param>
        private static List<HashSet<string>> GetTssWithVariedExpression(string cellLine, string[] RnaSources, bool useFullMap)
        {
            return RnaSources.Select(rnaSource => GetVariedElement("Tss", cellLine, rnaSource, useFullMap)).ToList();
        }

        /// <summary>
        /// Get drms with varied histone marks for all rna sources
        /// </summary>
        /// <returns>The drm with varied marks.</returns>
        /// <param name="cellLine">Cell line.</param>
        /// <param name="RnaSources">Rna sources.</param>
        /// <param name="useFullMap">Leave one out or all tissue sources</param>
        private static List<HashSet<string>> GetDrmWithVariedMarks(string cellLine, string[] RnaSources, bool useFullMap)
        {
            return RnaSources.Select(rnaSource => GetVariedElement("Drm", cellLine, rnaSource, useFullMap)).ToList();
        }


        /// <summary>
        /// Gets all drms within a specified range of a tss
        /// </summary>
        /// <returns>The drms in tss range.</returns>
        /// <param name="tss">Tss.</param>
        /// <param name="drmFile">Drm file.</param>
        /// <param name="range">Range.</param>
        /// <param name="drmWithVariedMarker">Range.</param>
        private static List<BedFile.Location> GetDrmsInTssRange(BedFile.Location tss, BedFile drmFile, int range, HashSet<string> drmWithVariedMarker)
        {
            return TRFScorer.EvaluateRange(
                drmFile.ChromosomeOrderedLocations [tss.Chromosome].Where(drm => drmWithVariedMarker.Contains(drm.Name)).ToList(), 
                tss.Start - range, 
                tss.End + range, 
                new List<BedFile.Location>(),
                (item, start) => item.End < start,
                (item, end) => item.Start <= end,
                (items, length, location) => 
                {
                items.Add(location);
                return items; 
            });
        }

        /// <summary>
        /// Get all drm-tss pairings for all rna sources in a cell line
        /// </summary>
        /// <returns>The all drms for variant expressed.</returns>
        /// <param name="cellLine">Cell line.</param>
        /// <param name="RnaSources">Rna sources.</param>
        /// <param name="tssData">Tss data.</param>
        /// <param name="drmFile">Drm file.</param>
        /// <param name="useFullMap">Leave one out or all tissue sources</param>
        private static List<ILookup<string, string>> GetAllDrmsForVariantExpressed(
            string cellLine, 
            string[] RnaSources, 
            List<Dictionary<string, BedFile.Location>> tssData, 
            BedFile drmFile,
            bool useFullMap)
        {
            List<ILookup<string, string>> allDrmMap = new List<ILookup<string, string>>();
            List<HashSet<string>> tssWithVariedExpression = GetTssWithVariedExpression(cellLine, RnaSources, useFullMap);
            
            for (int i = 0; i < RnaSources.Length; i++)
            {
                HashSet<string> drmWithVariedMarker = GetVariedElement("Drm", cellLine, RnaSources[i], useFullMap);

                allDrmMap.Add(tssWithVariedExpression[i]
                    .Where(tss => tssData[i].ContainsKey(tss))
                    .Select(tss => tssData[i][tss])
                    .Where(tss => drmFile.ChromosomeOrderedLocations.ContainsKey(tss.Chromosome))
                    .Select(tss => GetDrmsInTssRange(tss, drmFile, 1000000, drmWithVariedMarker)
                        .Select(y => new { Tss = tss.Name, Drm = y.Name } )
                    ).SelectMany(x => x)
                    .ToLookup(x => x.Tss, x => x.Drm));
            }

            return allDrmMap;
        }

        /// <summary>
        /// Gets NN map only using variably expressed or marked TSS or DRMs
        /// </summary>
        /// <returns>The NN map for variant expressed genes.</returns>
        /// <param name="cellLine">Cell line.</param>
        /// <param name="RnaSources">Rna sources.</param>
        /// <param name="drmFile">Drm file.</param>
        /// <param name="tssData">Tss data.</param>
        /// <param name="useFullMap">If set to <c>true</c> use full map.</param>
        private static List<ILookup<string, string>> GetNNMapForVariantExpressedGenes(
            string cellLine, 
            string [] RnaSources, 
            BedFile drmFile, 
            List<Dictionary<string, BedFile.Location>> tssData,
            bool useFullMap,
            bool forwardOnly)
        {
            //List<ILookup<string, string>> fullNNMaps = tssData.Select(x => CreateDrmNNMap(x, drmFile, forwardOnly)).ToList();
            List<ILookup<string, string>> filteredNNMaps = new List<ILookup<string, string>>();
            List<HashSet<string>> tssWithVariedExpression = GetTssWithVariedExpression(cellLine, RnaSources, useFullMap);

            for (int i = 0; i < RnaSources.Length; i++)
            {
                HashSet<string> drmWithVariedMarker = GetVariedElement("Drm", cellLine, RnaSources[i], useFullMap);

                var nnMap = CreateDrmNNMap(tssData[i], drmFile, tssWithVariedExpression[i], drmWithVariedMarker);
                
                filteredNNMaps.Add(nnMap);
            }

            return filteredNNMaps;
        }

        /// <summary>
        /// Multiples the regression data use xml config.
        /// </summary>
        /// <param name="xmlFile">Xml file.</param>
        /// <param name="dataSource">Data source.</param>
        /// <param name="cellLineOverride">Cell line override.</param>
        /// <param name="generateFeatures">If set to <c>true</c> generate features.</param>
        /// <param name="sampleDrms">If set to <c>true</c> sample drms.</param>
        /// <param name="useFullMap">If set to <c>true</c> use full map.</param>
        /// <param name="startRep">Start rep.</param>
        /// <param name="repCount">Rep count.</param>
        /// <param name="batch">Batch.</param>
        /// <param name="stage">Stage.</param>
        private static void MultipleRegressionDataUseXmlConfig(
            string xmlFile, 
            string dataSource,
            string cellLineOverride,
            bool generateFeatures,
            bool sampleDrms,
            bool useFullMap,
            int startRep,
            int repCount,
            int batch,
            RegressionStage stage)
        {
            if (!Config.DataSources.ContainsKey(dataSource))
            {
                throw new Exception(string.Format("Invalid data source: {0}\n Valid values:\n\t{1}", dataSource, string.Join("\n\t", Config.DataSources.Keys)));
            }

            Config config = new Config(xmlFile);

            if (startRep == default(int))
            {
                startRep = 0;
            }

            if (repCount == default(int))
            {
                repCount = 1;
            }

            string[] cellLines = config.CellLines;
            if (cellLineOverride != null)
            {
                cellLines = new string[] { cellLineOverride };
            }

            foreach (string cellLine in cellLines)
            {  
                Console.WriteLine("Loading drms for cell line {0}", cellLine);
                BedFile drmFile = new BedFile(config.GetDrmFileNameForCellLine(cellLine), BedFile.Bed4Drm);

                Console.WriteLine("Loading tss data for {0}", cellLine);
                List<Dictionary<string, BedFile.Location>> tssData = GetTssDataFromExpression(
                    config.RnaSources, 
                    i => config.CellLineTag.Replace(config.RnaSourceTag.Replace(config.ExpressionFileFormat, config.RnaSources[i]), cellLine));

                Console.WriteLine("Loading NN map for cell line and tss data");

                List<ILookup<string, string>> filteredNNMaps = GetNNMapForVariantExpressedGenes(cellLine, config.RnaSources, drmFile, tssData, useFullMap, false);

                MultipleRegression(
                    config,
                    cellLine,
                    dataSource,
                    drmFile,
                    tssData,
                    filteredNNMaps,
                    generateFeatures,
                    sampleDrms,
                    useFullMap,
                    startRep,
                    repCount,
                    batch,
                    stage);
            }
        }

        private static string GetDrmScoreFileName(string sourceFileName)
        {
            return GetScoreFileName(sourceFileName.Split('/').Last(), "DrmScores");
        }

        private static string GetTssScoreFileName(string sourceFileName)
        {
            return GetScoreFileName(sourceFileName.Split('/').Last(), "TssScores");
        }

        private static string GetScoreFileName(string sourceFileName, string scoreType)
        {
            return string.Format("../temp/YipMapRegressions/{0}.{1}.tsv", scoreType, sourceFileName);
        }

        private static Dictionary<string, BedFile.Location> UnionTssLocations(List<Dictionary<string, BedFile.Location>> tssData)
        {
            return tssData.Select(x => x)
                    .SelectMany(x => x)
                    .ToLookup(x => x.Key, x => x.Value)
                    .ToDictionary(x => x.Key, x => x.First());
        }

        private enum RegressionStage
        {
            GenerateFeatures,
            BuildTables,
            Regress,
            SampleDrms,
            Test
        }

        /// <summary>
        /// Does regression across thresholds of a given cell line and set of expression data
        /// </summary>
        /// <param name="config">Config.</param>
        /// <param name="cellLine">Cell line.</param>
        /// <param name="dataSource">Data source.</param>
        /// <param name="drmFile">Drm file.</param>
        /// <param name="tssData">Tss data.</param>
        /// <param name="variantExpressedNNMaps">Variant expressed NN maps.</param>
        /// <param name="generateFeatures">If set to <c>true</c> generate features.</param>
        /// <param name="sampleDrms">If set to <c>true</c> sample drms.</param>
        /// <param name="useFullMap">If set to <c>true</c> then do not leave the source cell line out of map generation</param>
        /// <param name="startRep">Start rep.</param>
        /// <param name="repCount">Rep count.</param>
        private static void MultipleRegression(
            Config config,
            string cellLine,
            string dataSource,
            BedFile drmFile,
            List<Dictionary<string, BedFile.Location>> tssData,
            List<ILookup<string, string>> variantExpressedNNMaps,
            bool generateFeatures,
            bool sampleDrms,
            bool useFullMap,
            int startRep,
            int repCount,
            int batch,
            RegressionStage stage)
        {
            int k = 0;
            int count = GetChIPDataSetNames(config.GetChipDataSourceFileList(dataSource, cellLine)).Count();

            Dictionary<string, BedFile.Location> universalGeneIds = UnionTssLocations(tssData);

            switch (stage)
            {
                case RegressionStage.GenerateFeatures:
                {
                    HashSet<string> usedTsses = new HashSet<string>();
                    HashSet<string> usedDrms = new HashSet<string>();
                    for (int i = 0; i < config.SourceCount; i++)
                    {
                        for (int j = 0; j < config.Thresholds.Length; j++)
                        {
                            GC.Collect();
                            
                            //Console.WriteLine("Processing map {0} on threshhold {1}", i, thresholds[j]);
                            
                            Dictionary<string, List<string>> drmMap;
                            Dictionary<string, List<string>> drmMapNN;
                            
                            LoadMaps(config.GetMapFileNames(cellLine)[i][j], tssData[i], drmFile, variantExpressedNNMaps[i], out drmMap, out drmMapNN);

                            usedTsses.UnionWith(drmMap.Keys);
                            usedTsses.UnionWith(drmMapNN.Keys);
                            usedDrms.UnionWith(drmMap.Values.SelectMany(x => x));
                            usedDrms.UnionWith(drmMapNN.Values.SelectMany(x => x));
                        }
                    }

                    // Dictionary<string, int> drmIndices = 
                    TRFScorer.CreateIndex(drmFile.Locations, "DRM", cellLine);
                    // Dictionary<string, int> tssIndices = 
                    TRFScorer.CreateIndex(universalGeneIds, "TSS", cellLine);

                    foreach (string filename in GetChIPDataFileNames(config.GetChipDataSourceFileList(dataSource, cellLine)))
                    {
                        var chipFile = new BedFile(filename, filename.Trim().Contains("broad") ? BedFile.Bed6Plus3Layout : BedFile.Bed6Plus4Layout);

                        string baseFileName = chipFile.FileName.Split('.').First();
                        Console.WriteLine("\tScoring feature {2}: {0} of {1}", k, count, baseFileName);

                        TRFScorer.WriteScoreNameValuePairs(GetDrmScoreFileName(baseFileName), TRFScorer.CreateDrmScores(drmFile, chipFile, false), null);
                        TRFScorer.WriteScoreNameValuePairs(GetTssScoreFileName(baseFileName), TRFScorer.CreateTssScores(universalGeneIds, chipFile, config.D0, config.Range), null);

                        GC.Collect();
                        k++;
                    }

                    return;
               }
            }

            Func<string, string> MakeMapPromOnlyScript = tableName => tableName.Replace(".tsv", ".PromOnly.R");
            Func<string, string> MakeMapScript = tableName => tableName.Replace(".tsv", ".TssAndDrm.R");
            Func<string, string> MakeMapNNScript = tableName => tableName.Replace(".tsv", ".TssAndNNDrm.R");
            Func<string, string, string> MakeSampledPromOnlyScript = (tableName, theta) => tableName
                .Replace(".1NNR.", "." + theta + "NNR.")
                .Replace(".tsv", ".PromOnly.R");
            Func<string, string, string> MakeSampledNNScript = (tableName, theta) => tableName
                .Replace(".1NNR.", "." + theta + "NNR.")
                .Replace(".tsv", ".TssAndNNDrm.R");

            Func<string, string> MakeNNTableName = mapTableName => mapTableName.Replace(".tsv", "NN.tsv");


            switch (stage)
            {
                case RegressionStage.BuildTables:
                {
                    List<string> chipDataSets = GetChIPDataSetNames(config.GetChipDataSourceFileList(dataSource, cellLine)).ToList();
                    
                    var drmIndex = TRFScorer.ReadIndex("DRM", cellLine);
                    var tssIndex = TRFScorer.ReadIndex("TSS", cellLine);
                    
                    List<double[]> drmScores = new List<double[]>();
                    List<double[]> tssScores = new List<double[]>();
                    foreach (string chipFile in chipDataSets)
                    {
                        Console.WriteLine("\tReading feature {2}: {0} of {1}", k++, chipDataSets.Count, GetDrmScoreFileName(chipFile));
                        drmScores.Add(TRFScorer.ReadScoreNameValuePairs<double>(drmIndex, GetDrmScoreFileName(chipFile), x => double.Parse(x)));
                        tssScores.Add(TRFScorer.ReadScoreNameValuePairs<double>(tssIndex, GetTssScoreFileName(chipFile), x => double.Parse(x)));
                        
                        GC.Collect();
                    }

                    for (int i = 0; i < config.SourceCount; i++)
                    {
                        string nnSampledMapOutFile = config.GetTableFileNames(cellLine)[i][0].Replace(".tsv", "NNR.tsv");

                        Dictionary<string, double> expression = GetExpression(config.GetExpressionFileName(config.RnaSources[i], cellLine));

                        ILookup<string, int> ExpressionTSSLocations = GetTSSMap(expression.Keys);
                        Dictionary<string, string> ChIPToExpressionTSS_NN = NearestTSS(variantExpressedNNMaps[i].Select(x => x.Key), ExpressionTSSLocations);

                        var fullMapDict = variantExpressedNNMaps[i].Select(x => new { Tss = x.Key, Drms = new List<string>(x) } ).ToDictionary(x => x.Tss, x => x.Drms);
                        R.CreateTable(fullMapDict, tssIndex, tssScores, drmIndex, drmScores, expression, ChIPToExpressionTSS_NN, nnSampledMapOutFile, chipDataSets);

                        for (int j = 0; j < config.Thresholds.Length; j++)
                        {
                            Dictionary<string, List<string>> drmMap;
                            Dictionary<string, List<string>> drmMapNN;
                            Dictionary<string, List<string>> drmMapAllDrms = new Dictionary<string, List<string>>();
                            
                            LoadMaps(config.GetMapFileNames(cellLine)[i][j], tssData[i], drmFile, variantExpressedNNMaps[i], out drmMap, out drmMapNN);

                            if (drmMap.Count >= tssScores.Count + drmScores.Count)
                            {
                                string drmMapOutFile = config.GetTableFileNames(cellLine)[i][j];
                                string nnMapOutFile = drmMapOutFile.Replace(".tsv", "NN.tsv");
                                string mapTssAllDrmOutFile = drmMapOutFile.Replace(".tsv", "DS.tsv");

                                Dictionary<string, string> ChIPToExpressionTSS = NearestTSS(drmMap.Select(x => x.Key), ExpressionTSSLocations);

                                R.CreateTable(drmMap, tssIndex, tssScores, drmIndex, drmScores, expression, ChIPToExpressionTSS, drmMapOutFile, chipDataSets);
                                R.CreateTable(drmMapNN, tssIndex, tssScores, drmIndex, drmScores, expression, ChIPToExpressionTSS, nnMapOutFile, chipDataSets);
                            }
                        }
                    }
                    break;
                }

                case RegressionStage.Regress:
                {
                    for (int i = 0; i < config.SourceCount; i++)
                    {
                        List<string> chipDataSets = GetChIPDataSetNames(config.GetChipDataSourceFileList(dataSource, cellLine)).ToList();

                        var tssColumns = R.MakeTssColumnNames(chipDataSets);
                        var drmColumns = R.MakeDrmColumnNames(chipDataSets);

                        Console.WriteLine("<table>");
                        Console.WriteLine(trhs + "Threshold" + th + "TssCount" +
                                          th + "MapTssPromOnly" + th + "MapTssAndDrm" + th + "MapTssAndNNDrm" + 
                                          th + "SampledTssPromOnly" + th + "SampledTssAndNNDrm" +
                                          th + "SampledTssPromOnlyDev" + th + "SampledTssAndNNDrmDev" + 
                                          th + "SampledTssImprovementDev" + trhe);
                        
                        string nnSampledMapOutFile = config.GetTableFileNames(cellLine)[i][0].Replace(".tsv", "NNR.tsv");

                        //Dictionary<string, double> expression = GetExpression(config.GetExpressionFileName(config.RnaSources[i], cellLine));
                        
                        //ILookup<string, int> ExpressionTSSLocations = GetTSSMap(expression.Keys);
                        //Dictionary<string, string> ChIPToExpressionTSS_NN = NearestTSS(variantExpressedNNMaps[i].Select(x => x.Key), ExpressionTSSLocations);
                        
                        //var fullMapDict = variantExpressedNNMaps[i].Select(x => new { Tss = x.Key, Drms = new List<string>(x) } ).ToDictionary(x => x.Tss, x => x.Drms);
                        //R.CreateTable(fullMapDict, tssIndex, tssScores, drmIndex, drmScores, expression, ChIPToExpressionTSS_NN, nnSampledMapOutFile, chipDataSets);



                        for (int j = 0; j < config.Thresholds.Length; j++)
                        {
                            Dictionary<string, List<string>> drmMap = LoadMap(config.GetMapFileNames(cellLine)[i][j], tssData[i])
                                .Select(x => new { Key = x.Key, Value = new List<string>(x) })
                                .ToDictionary(x => x.Key, x => x.Value);

                            if (drmMap.Count >= tssColumns.Count() + drmColumns.Count())
                            {
                                string drmMap_TableFile = config.GetTableFileNames(cellLine)[i][j];
                                string nnMap_TableFile = MakeNNTableName(drmMap_TableFile);
                                //string mapTss_AllDrmTableFile = drmMap_TableFile.Replace(".tsv", "DS.tsv");

                                // Map TSS regression for promoter only and assigned DRMs

                                string drmMap_ScriptFile_PromOnly = MakeMapPromOnlyScript(drmMap_TableFile);
                                string drmMap_ScriptFile_TssDrm = MakeMapScript(drmMap_TableFile);

                                double[] mapTss_Rsq_PromOnly = R.RegressColumns(
                                    drmMap_ScriptFile_PromOnly,
                                    drmMap_TableFile,
                                    tssColumns,
                                    R.RegressColumnsParam.AllRows,
                                    repCount,
                                    batch);

                                double[] mapTss_Rsq_Drm = R.RegressColumns(
                                    drmMap_ScriptFile_TssDrm,
                                    drmMap_TableFile, 
                                    tssColumns.Concat(drmColumns),
                                    R.RegressColumnsParam.AllRows,
                                    repCount,
                                    batch);

                                // NN DRM Regression on Map TSSes

                                string nnMap_ScriptFile_TssDrm = MakeMapNNScript(nnMap_TableFile); 

                                double[] mapTss_Rsq_NNDrm = R.RegressColumns(
                                    nnMap_ScriptFile_TssDrm,
                                    nnMap_TableFile,
                                    tssColumns.Concat(drmColumns),
                                    R.RegressColumnsParam.AllRows,
                                    repCount,
                                    batch);

                                // Sampled tss prom only and nn regression

                                string nnSampled_ScriptFile_PromOnly = MakeSampledPromOnlyScript(nnSampledMapOutFile, config.Thresholds[j]);
                                    
                                double[] sampledTss_Rsq_PromOnly = R.RegressColumns(
                                    nnSampled_ScriptFile_PromOnly,
                                    nnSampledMapOutFile,
                                    tssColumns,
                                    drmMap.Count,
                                    repCount,
                                    batch);
                                                                
                                string nnSampled_ScriptFile_TssDrm = MakeSampledPromOnlyScript(nnSampledMapOutFile, config.Thresholds[j]);

                                double[] sampledTss_Rsq_NNDrm = R.RegressColumns(
                                    nnSampled_ScriptFile_TssDrm,
                                    nnSampledMapOutFile,
                                    tssColumns.Concat(drmColumns),
                                    drmMap.Count,
                                    repCount,
                                    batch);

                                // Write R2 values for the batch
                                Console.WriteLine(trds + config.Thresholds[j] + td + drmMap.Count +
                                                  td + mapTss_Rsq_PromOnly.Average() + td + mapTss_Rsq_Drm.Average() + td + mapTss_Rsq_NNDrm.Average() + 
                                                  td + sampledTss_Rsq_PromOnly.Average() + td + sampledTss_Rsq_NNDrm.Average() + 
                                                  td + Stats.StdDev(sampledTss_Rsq_PromOnly) + td + Stats.StdDev(sampledTss_Rsq_NNDrm) + 
                                                  td + Stats.StdDev(sampledTss_Rsq_NNDrm.Zip(sampledTss_Rsq_PromOnly, (x, y) => x - y)) + trde);
                            }
                        }

                        Console.WriteLine("</table>");
                    }
                    break;
                }

                case RegressionStage.SampleDrms:
                {
                    //HashSet<string> drmWithVariedMarker;
                    //using (TextReader tr = new StreamReader(string.Format("Maps/InHouse/Correlation/{0}.All.Drm.{1}.nsv", cellLine, config.RnaSources[i])))
                    //{
                    //    drmWithVariedMarker = new HashSet<string>(tr.ReadToEnd().Split('\n'));
                    //}

                    //if (sampleDrms)
                    //{
                    //   drmMapAllDrms = drmMap.Where(mapEntry => tssData[i].ContainsKey(mapEntry.Key))
                    //        .Select(mapEntry => new 
                    //                {
                    //            Tss = mapEntry.Key,
                    //            Drms = GetDrmsInTssRange(tssData[i][mapEntry.Key], drmFile, 1000000, drmWithVariedMarker)
                    //        })
                    //            .ToDictionary(x => x.Tss, x => x.Drms.Select(y => y.Name).ToList());
                    //}
                    break;
                }

                case RegressionStage.Test:
                {
                    Console.WriteLine("<h1>" + cellLine + "</h1>");
                    for (int i = 0; i < config.SourceCount; i++)
                    {
                        Console.WriteLine("<h2>" + config.RnaSources[i] + "</h2>\n<table>");
                        string nnSampledMapOutFile = config.GetTableFileNames(cellLine)[i][0].Replace(".tsv", "NNR.tsv");

                        Console.WriteLine(trhs + "Theta" + th + "Map vs Prom" + th + "Map vs NNMap" + th + "NNMap vs Prom" + th + 
                                          "SNNMap vs SProm" + th + "Map vs SNNMap" + th + "NNMap vs SNNMap" + trhe);

                        for (int j = 0; j < config.Thresholds.Length; j++)
                        {
                            string drmMap_TableFile = config.GetTableFileNames(cellLine)[i][j];
                            string nnMap_TableFile = MakeNNTableName(drmMap_TableFile);

                            if (File.Exists(drmMap_TableFile))
                            {
                                string drmMap_ScriptFile_PromOnly = MakeMapPromOnlyScript(drmMap_TableFile);
                                string drmMap_ScriptFile_TssDrm = MakeMapScript(drmMap_TableFile);
                                string nnMap_ScriptFile_TssDrm = MakeMapNNScript(nnMap_TableFile); 
                                string nnSampled_ScriptFile_PromOnly = MakeSampledPromOnlyScript(nnSampledMapOutFile, config.Thresholds[j]);
                                string nnSampled_ScriptFile_TssDrm = MakeSampledPromOnlyScript(nnSampledMapOutFile, config.Thresholds[j]);

                                var mapPromOnlyR2 = R.ReadR2Set(drmMap_ScriptFile_PromOnly, batch);
                                var mapR2 = R.ReadR2Set(drmMap_ScriptFile_TssDrm, batch);
                                var nnMapR2 = R.ReadR2Set(nnMap_ScriptFile_TssDrm, batch);
                                var sampledNNPromOnlyR2 = R.ReadR2Set(nnSampled_ScriptFile_PromOnly, batch);
                                var sampledNNR2 = R.ReadR2Set(nnSampled_ScriptFile_TssDrm, batch);

                                //repCount = 100;
                                var pvalMap = Stats.StatisticalTests.PermutationTest(mapR2, mapPromOnlyR2, repCount);
                                var pvalNNMap = Stats.StatisticalTests.PermutationTest(nnMapR2, mapPromOnlyR2, repCount);
                                var pvalMapNNMap = Stats.StatisticalTests.PermutationTest(mapR2, nnMapR2, repCount);
                                var pvalSampledNN = Stats.StatisticalTests.PermutationTest(sampledNNR2, sampledNNPromOnlyR2, repCount);
                                var pvalMapSampledNN = Stats.StatisticalTests.PermutationTest(mapR2, sampledNNR2, repCount);
                                var pvalNNMapSampledNN = Stats.StatisticalTests.PermutationTest(nnMapR2, sampledNNR2, repCount);

                                Console.WriteLine(trds + config.Thresholds[j] + td + pvalMap + td + pvalMapNNMap + td + pvalNNMap + td + 
                                                  pvalSampledNN + td + pvalMapSampledNN + td + pvalNNMapSampledNN + trde);

                            }
                        }
                        Console.WriteLine("</table>");
                    }
                    break;
                }
            }









            /*for (int i = 0; i < config.SourceCount; i++)
            {
                Dictionary<string, double> expression = GetExpression(config.GetExpressionFileName(config.RnaSources[i], cellLine));

                Console.WriteLine("Make expression tss location lookup");
                ILookup<string, int> ExpressionTSSLocations = GetTSSMap(expression.Keys);

                Dictionary<string, string> ChIPToExpressionTSS_NN = NearestTSS(variantExpressedNNMaps[i].Select(x => x.Key), ExpressionTSSLocations);

                HashSet<string> drmWithVariedMarker;
                
                using (TextReader tr = new StreamReader(string.Format("Maps/InHouse/Correlation/{0}.All.Drm.{1}.nsv", cellLine, config.RnaSources[i])))
                {
                    drmWithVariedMarker = new HashSet<string>(tr.ReadToEnd().Split('\n'));
                }

                var tssColumns = R.MakeTssColumnNames(chipDataSets);
                var drmColumns = R.MakeDrmColumnNames(chipDataSets);

                Console.WriteLine(trhs + "Threshold" + th + "TssCount" +
                                  th + "MapTssPromOnly" + th + "MapTssAndDrm" + th + "MapTssAndNNDrm" + 
                                  th + "SampledTssPromOnly" + th + "SampledTssAndNNDrm" + trhe);

                string nnSampledMapOutFile = config.GetOutputFileNames(cellLine)[i][0].Replace(".tsv", "NNR.tsv");

                var fullMapDict = variantExpressedNNMaps[i].Select(x => new { Tss = x.Key, Drms = new List<string>(x) } ).ToDictionary(x => x.Tss, x => x.Drms);
                R.CreateTable(fullMapDict, tssIndex, tssScores, drmIndex, drmScores, expression, ChIPToExpressionTSS_NN, nnSampledMapOutFile, chipDataSets);

                for (int j = 0; j < config.Thresholds.Length; j++)
                {
                    GC.Collect();

                    Console.WriteLine("Processing map {0} on threshhold {1}", i, config.Thresholds[j]);
                    Console.Error.WriteLine("Processing map {0} on threshhold {1}", i, config.Thresholds[j]);

                    Dictionary<string, List<string>> drmMap;
                    Dictionary<string, List<string>> drmMapNN;
                    Dictionary<string, List<string>> drmMapAllDrms = new Dictionary<string, List<string>>();

                    LoadMaps(config.GetMapFileNames(cellLine)[i][j], tssData[i], drmFile, variantExpressedNNMaps[i], out drmMap, out drmMapNN);

                    if (sampleDrms)
                    {
                        drmMapAllDrms = drmMap.Where(mapEntry => tssData[i].ContainsKey(mapEntry.Key))
                            .Select(mapEntry => new 
                            {
                                Tss = mapEntry.Key,
                                Drms = GetDrmsInTssRange(tssData[i][mapEntry.Key], drmFile, 1000000, drmWithVariedMarker)
                            })
                            .ToDictionary(x => x.Tss, x => x.Drms.Select(y => y.Name).ToList());
                    }
                                                                                                                                                        ;

                    Dictionary<string, string> ChIPToExpressionTSS = NearestTSS(drmMap.Select(x => x.Key), ExpressionTSSLocations);

                    string drmMapOutFile = config.GetOutputFileNames(cellLine)[i][j];
                    string nnMapOutFile = drmMapOutFile.Replace(".tsv", "NN.tsv");
                    string mapTssAllDrmOutFile = drmMapOutFile.Replace(".tsv", "DS.tsv");

                    if (drmMap.Count >= tssScores.Count + drmScores.Count)
                    {
                        nTss[i][j] = R.CreateTable(drmMap, tssIndex, tssScores, drmIndex, drmScores, expression, ChIPToExpressionTSS, drmMapOutFile, chipDataSets);

                        double[] mapTssPromOnly = R.RegressColumns(drmMapOutFile, tssColumns, -1, repCount, drmMapOutFile.Replace(".tsv", ".PromOnly.R"));
                        double[] mapTssAndDrm = R.RegressColumns(drmMapOutFile, tssColumns.Concat(drmColumns), -1, repCount, drmMapOutFile.Replace(".tsv", ".TssAndDrm.R"));

                        R.CreateTable(drmMapNN, tssIndex, tssScores, drmIndex, drmScores, expression, ChIPToExpressionTSS, nnMapOutFile, chipDataSets);

                        double[] mapTssAndNNDrm = R.RegressColumns(nnMapOutFile, tssColumns.Concat(drmColumns), -1, repCount, nnMapOutFile.Replace(".tsv", ".TssAndNNDrm.R"));

                        double[] sampledTssPromOnly = R.RegressColumns(nnSampledMapOutFile, tssColumns, nTss[i][j], repCount, nnSampledMapOutFile.Replace(".tsv", ".PromOnly.R"));
                        double[] sampledTssAndNNDrm = R.RegressColumns(nnSampledMapOutFile, tssColumns.Concat(drmColumns), nTss[i][j], repCount, nnSampledMapOutFile.Replace(".tsv", ".TssAndNNDrm.R"));

                        //List<string> batchFiles = new List<string>();
                        //string[] outFileNames = { drmMapOutFile, nnMapOutFile, nnSampledMapOutFile, mapTssAllDrmOutFile };

                        /*if (sampleDrms)
                        {
                            batchFiles = R.CreateTableAndList(
                                drmMap, drmMapAllDrms, 
                                tssIndex, tssScores, 
                                drmIndex, drmScores, 
                                expression, ChIPToExpressionTSS, 
                                mapTssAllDrmOutFile, chipDataSets, repCount);
                        }*




                        //double[] yipMapScores =

                        //double[] Rsquareds;
                        //double[][] pvals;
                        //string scriptBaseName = drmMapOutFile.Replace(".tsv", ".R");
                        //R.BootRegressMapScoreData(outFileNames, batchFiles, chipDataSets, repCount, scriptBaseName, out Rsquareds, out pvals);

                        //allRsquareds[i][j] = Rsquareds;
                        //allPvals[i][j] = pvals;
                        //Console.Error.WriteLine("\t" + Rsquareds[0] + "\t" + Rsquareds[3] + "\t" + Rsquareds[6] + "\t" + Rsquareds[9]);

                        //double[] nnMapScores = R.RegressMapScoreData(nnMapOutFile, chipDataSets, R.RegressionFunction.Linear);

                        //double[] nnRandomMapScores = R.RegressMapScoreData(nnMapOutFile, chipDataSets, R.RegressionFunction.Linear);


                        // Bootstrap sampling from within C#
                        /*List<string> unsampledTssNames = new List<string>(tssNames);
                        var sampledTssNames = drmMap.Select(x => 
                        {
                            int sample = rng.Next(unsampledTssNames.Count);
                            string tss = unsampledTssNames[sample];
                            unsampledTssNames.RemoveAt(sample);
                            return tss;
                        });

                        Dictionary<string, List<string>> drmRandomNNTsses = sampledTssNames
                            .Select(x => new { Tss = x, Drms = fullNNMaps[i][x].ToList() })
                            .ToDictionary(x => x.Tss, x => x.Drms);*/


                        /*Console.WriteLine("NN random scores:\t{0}", string.Join("\t", nnRandomMapScores));
                   
                        allYipScores[i][j] = yipMapScores;
                        allNNScores[i][j] = nnMapScores;*
                    }
                    else
                    {
                        Console.Error.WriteLine("Skipping threshold {0} on map {1} due to insufficient map size", config.Thresholds[j], config.RnaSources[i]);  
                        double[] Rsquareds = new double[10].Select(x => double.NaN).ToArray();
                        double[][] pvals = new double[10].Select(x => new double[10].Select(y => double.NaN).ToArray()).ToArray();

                        allRsquareds[i][j] = Rsquareds;
                        allPvals[i][j] = pvals;
                    }
                }
            }

            /*string[] dataSets = { "Map", "NN", "RandomNN", "Map TSS/Sampled DRM" };
            string[] subset = { "DRM+TSS", "TSS", "DRM" };
            
            Func<int, string> DataSet = (index) => dataSets[index / 3];
            Func<int, string> DataSubset = (index) => subset[index % 3];

            int[] indices = { 0, 3, 6, 2, 5, 8, 1, 7, 9 };
            /// int[] DRMTSSindices = { 0, 3, 6 };
            /// int[] DRMindices = { 2, 5, 8 };
            /// int[] TSSindices = { 1, 7 };

            List<string> dataSetFullNames = indices.Select(x => string.Format("{0} {1}", DataSubset(x), DataSet(x))).ToList();

            for (int i = 0; i < config.SourceCount; i++)
            {
                Console.WriteLine("<table>");
                Console.WriteLine("<tr><th>-Log10 Threshhold</th><th colspan=10 align='center'>{0} Expression in {1}</th></tr>", config.RnaSources[i], cellLine);

                Console.WriteLine("<tr><th></th><th>TSS Count</th><th>" + string.Join("</th><th>", dataSetFullNames) + "</th></tr>");

                for (int j = 0; j < config.Thresholds.Length; j++)
                {
                    Console.WriteLine("<tr><td>{0}</td>", config.Thresholds[j]);
                    Console.WriteLine(
                        "<td>{0}</td><td>{1}</td>", 
                        nTss[i][j],
                        string.Join("</td><td>", indices.Select(x => string.Format("{0:F3}", allRsquareds[i][j][x])))
                        );
                    Console.WriteLine("</tr>");
                }
                Console.WriteLine("</table>");

            }

            for (int i = 0; i < config.SourceCount; i++)
            {
                Console.WriteLine("\n\n\n<table>");
                for (int j = 0; j < config.Thresholds.Length; j++)
                {
                    Console.WriteLine("<tr><td><h4>{0} {1} threshold {2}</h4>", config.RnaSources[i], cellLine, config.Thresholds[j]);
                    Console.WriteLine("<table>");

                    Console.WriteLine("<tr><td></td><td>{0}</td></tr>", string.Join("</td><td>", dataSetFullNames));

                    for (int l = 0; l < indices.Length; l++)
                    {
                        Console.WriteLine("<tr><td>{0}</td><td>{1}</td></tr>",
                            dataSetFullNames[l],
                            string.Join("</td><td>", indices.Select(x => allPvals[i][j][indices[l]][x])));
                    }

                    Console.WriteLine("</td></tr></table>");
                }
                Console.WriteLine("</table>");

            }*/
        }

        private static Regex stripReplicate = new Regex("Rep[0-9]");
        private static Regex stripExtension = new Regex("StdPk.*");

        private static Dictionary<string, string> GetHistoneModFilesForCellLine(string cellLine, string histoneRoot)
        {
            string histoneDirectory = string.Format("{0}/{1}", histoneRoot, cellLine);
            string[] histoneFiles = Directory.GetFiles(histoneDirectory, "*Peak");
            
            Regex stripHeader = new Regex(string.Format(".*{0}", cellLine));
            
            return histoneFiles.Select(x => new 
                                                                  {
                Mod = stripExtension.Replace(stripHeader.Replace(stripReplicate.Replace(x, ""), ""), ""),
                File = x
            }).ToLookup(x => x.Mod, x => x.File)
            .ToDictionary(x => x.Key, x => x.First());  // pick only one file when more than one replicate occurs
        }

        private static bool IsTwoFoldChange(IEnumerable<double> data)
        {
            double mean = data.Average();
            double min = data.Min();
            double max = data.Max();

            if (mean == 0)
            {
                return false;
            }

            if (max - mean >= 2 * mean)
            {
                return true;
            }

            if (mean == min)
            {
                return false;
            }

            if (mean - min >= mean / 3)
            {
                return true;
            }

            return false;
        }

        /// <summary>
        /// Builds a correlation map for a given cell line's drms. 
        /// Optionally omits source line from data.
        /// Allows command line file to be provided.
        /// </summary>
        /// <param name="configFile">Config file.</param>
        /// <param name="cellLine">Cell line.</param>
        /// <param name="bedFileName">Bed file of DRMs</param>
        /// <param name="omitDrmSource">If set to <c>true</c> omit drm source.</param>
        private static void BuildMap(
            string configFile, 
            string cellLine,
            string bedFileName,
            bool omitDrmSource)
        {
            XDocument doc = XDocument.Load(configFile);
                                          
            string histoneRoot = doc.Descendants("HistoneRoot").First().Value.ToString();
            string expressionRoot = doc.Descendants("ExpressionRoot").First().Value.ToString();
            string[] cellLines = doc.Descendants("CellLines").First().Value.ToString().Split(',')
                .Where(cl => !string.IsNullOrEmpty(cellLine) && 
                       ((omitDrmSource && string.Compare(cl, cellLine) != 0) || !omitDrmSource)).ToArray();
           
            string[] expressionTypes = doc.Descendants("ExpressionTypes").First().Value.ToString().Split(',');

            string cellLineDrmFileName = string.IsNullOrEmpty(bedFileName) ? 
                string.Format(doc.Descendants("CellLineDrmFileNames").First().Value.ToString(), cellLine) :
                bedFileName;

            int minFeatureCount = string.IsNullOrEmpty(cellLine) ? 7 : 6;

            BedFile drmFile = new BedFile(cellLineDrmFileName, BedFile.Bed4Drm);
        
            Console.WriteLine("Building drm-tss pairs ...");
           
            // Expression data by type and cell line
            Dictionary<string, Dictionary<string, string>> validCellLineExpression = cellLines
                .Select(cl => expressionTypes
                        .Select(et => new
                        {
                            CellLine = cl,
                            ExpType = et,
                            Filename = string.Format("{0}/{1}/{2}.tsv", expressionRoot, cl, et)
                        })
                        .Where(expFile => File.Exists(expFile.Filename)))
                    .SelectMany(x => x)
                    .ToLookup(x => x.ExpType, x => x)
                    .Where(x => x.Count() >= minFeatureCount) // Ensure at least seven cell lines have the type of expression
                    .ToDictionary(x => x.Key, x => x.ToDictionary(y => y.CellLine, y => y.Filename));
            
            // Histone data by type and cell line
            Dictionary<string, Dictionary<string, string>> validCellLineHistoneMods = cellLines
                .Select(cl => GetHistoneModFilesForCellLine(cl, histoneRoot)
                        .Select(hm => new
                        {
                    CellLine = cl,
                    ModName = hm.Key,
                    ModFile = hm.Value
                }))
                    .SelectMany(x => x)
                    .ToLookup(x => x.ModName, x => x)
                    .Where(x => x.Count() >= minFeatureCount) // Ensure at least seven cell lines have the modification
                    .ToDictionary(x => x.Key, x => x.ToDictionary(y => y.CellLine, y => y.ModFile));

            foreach (var expressionType in validCellLineExpression)
            { 
                Console.WriteLine("Loading {0} expression data ...", expressionType.Key);

                // Expression data by TSS and cell line
                var rawExpData = expressionType.Value
                    .Select(cl => 
                    {
                        using (TextReader tr = new StreamReader(cl.Value))
                        {
                            Console.WriteLine("\t\tReading file {0}", cl.Value);
                            return tr.ReadToEnd()
                                .Split('\n')
                                .Where(line => !string.IsNullOrWhiteSpace(line))
                                .Select(line => 
                                {
                                    string[] fields = line.Split('\t');
                                    return new 
                                    {
                                        CellLine = cl.Key,
                                        Tss = fields [0],
                                        Value = double.Parse(fields [1]),
                                        Chromosome = fields[2]
                                    };
                                });
                        }
                    })
                    .SelectMany(x => x)
                    .Where(x => x.Value > 0)
                    .ToLookup(x => x.Tss, x => x)
                    .Where(x => x.Count() > 0)
                    .Select(x => new 
                    {
                        Tss = x.Key,
                        Chromosome = x.First().Chromosome,
                        Position = int.Parse(x.Key.Split('_')[1]),
                        Expression = x.ToLookup(y => y.CellLine, y => y.Value)
                          .ToDictionary(y => y.Key, y => y.Max())
                    });

                Console.WriteLine("\t\tBuilding expression lookup");
                Dictionary<string, Dictionary<string, double>> expData = rawExpData.ToDictionary(x => x.Tss, x => x.Expression);
                // Only select tsses where we have at least the minimum number of data points, minimum expression and minimum fold change
                Console.WriteLine("\t\tBuilding valid expression tss set");
                var tssData = rawExpData.Where(x => x.Expression.Count() >= minFeatureCount &&
                                               x.Expression.Count(y => y.Value >= 2) > 0)
                    .ToDictionary(x => x.Tss, x => new { Chromosome = x.Chromosome, Position = x.Position });

                Console.WriteLine("\t\t\t{0} valid tsses", tssData.Count());

                Console.WriteLine("Getting DRM-TSS pairs");
                Dictionary<string, List<BedFile.Location>> tssDrmMapTmp = tssData
                    .Where(tss => drmFile.ChromosomeOrderedLocations.ContainsKey(tss.Value.Chromosome))
                    .Select(tss => new 
                    {
                        Tss = tss.Key,
                        Drms = TRFScorer.EvaluateRange(
                            drmFile.ChromosomeOrderedLocations [tss.Value.Chromosome], 
                            tss.Value.Position - 1000000, 
                            tss.Value.Position + 1000000, 
                            new List<BedFile.Location>(),
                            (item, start) => item.End < start,
                            (item, end) => item.Start <= end,
                            (items, length, location) => 
                            {
                                items.Add(location);
                                return items; 
                            })
                    }).ToDictionary(x => x.Tss, x => x.Drms);

                Dictionary<string, List<BedFile.Location>> tssDrmMap = tssDrmMapTmp
                    .Select(x => new 
                            {
                        Tss = x.Key,
                        Drms = x.Value
                           .Where(y => Math.Abs(tssData[x.Key].Position - y.Start) > 10000 && Math.Abs(tssData[x.Key].Position - y.End) > 10000)
                           .ToList()
                    })
                    .Where(x => x.Drms.Count > 0)
                    .ToDictionary(x => x.Tss, x => x.Drms);

                List<string> allTssDrmPairs = new List<string>(); // TSS#DRM string values of considered pairs

                foreach (var histoneType in validCellLineHistoneMods)
                {
                    Console.WriteLine("\tLoading histone data {0} ...", histoneType.Key);

                    // Histone data by DRM and cell line
                    Dictionary<string, Dictionary<string, double>> histData = histoneType.Value
                        .Select(cl =>
                        {
                            BedFile peaks = new BedFile(
                                cl.Value, 
                                cl.Value.Contains("broad") ? BedFile.Bed6Plus3Layout : BedFile.Bed6Plus4Layout);

                            return drmFile.Locations
                                .Where(x => peaks.ChromosomeOrderedLocations.ContainsKey(x.Value.Chromosome))
                                .Select(x => new
                                {
                                    CellLine = cl.Key,
                                    Drm = x.Key,
                                    Value = TRFScorer.EvaluateRange<double>(
                                        peaks.ChromosomeOrderedLocations[x.Value.Chromosome],
                                        x.Value.Start,
                                        x.Value.End,
                                        0,
                                        (item, start) => item.End < start,
                                        (item, end) => item.Start <= end,
                                        (aggregateScore, length, peak) => 
                                        {
                                            double d = 0.0;
                                            if (double.TryParse(peak.Data [peaks.FileLayout.SignalValue], out d))
                                            {
                                                // Get max peak value in range
                                                return d > aggregateScore ? d : aggregateScore;
                                            }
                                            return aggregateScore;
                                        })
                                });
                        })
                        .SelectMany(x => x)
                        .ToLookup(x => x.Drm, x => x)
                        .ToDictionary(x => x.Key, x => x.ToDictionary(y => y.CellLine, y => y.Value));

                    Console.WriteLine("\t\t\tBuilding correlation data");
                    int k = 0;
                    //string lastTss = "";
                    var rawcordata = tssDrmMap.Where(tss => expData.ContainsKey(tss.Key))
                        .Select(tss => 
                        {
                            Dictionary<string, double> tssExpression = expData[tss.Key];
                            return tss.Value.Where(drm => histData.ContainsKey(drm.Name))
                                .Select(drm => 
                                {
                                    var v = histData[drm.Name].Where(data => tssExpression.ContainsKey(data.Key));
                                    return new
                                    {
                                        Tss = tss.Key,
                                        Drm = drm.Name,
                                        VectorCsv = new Tuple<string, string>(string.Join(", ", v.Select(data => data.Value)),
                                                                              string.Join(", ", v.Select(data => tssExpression[data.Key]))),
                                        Vector = v.Select(data => new Stats.Point2 { x = data.Value, y = tssExpression[data.Key] })
                                    };
                                })
                                .Where(pair => pair.Vector.Count() >= minFeatureCount && //pair.Vector.Count(x => x.y >= 2) > 0 &&
                                       pair.Vector.Count(point => point.x > 0) > 0 &&
                                       pair.Vector.Count(point => point.y >= 2) > 0 && 
                                       IsTwoFoldChange(pair.Vector.Select(point => point.y)))
                                .ToList()
                                .Select(pair => 
                                {
                                    double correlation = Stats.Pearson(pair.Vector);
                                    return new
                                    {
                                        Tss = pair.Tss,
                                        Drm = pair.Drm,
                                        Correlation = correlation,
                                        Pvalue = 2 * Stats.Phi(- Math.Abs(Stats.FisherTransformZScore(correlation, pair.Vector.Count())))
                                    };
                                });
                        })
                        .SelectMany(x => 
                        {
                            k++;
                            if (k % 10000 == 0)
                            {
                                Console.WriteLine("\t\t\t\tProcessed {0} tsses", k);
                            }
                            return x;
                        });

                    double count = rawcordata.Count();
                    var cordata = rawcordata
                        .Select(x => new { Tss = x.Tss, Drm = x.Drm, Correlation = x.Correlation, Pvalue = x.Pvalue * count } )
                        .ToList();

                    Console.WriteLine("########");
                    Console.WriteLine("\t### {0} tss-drm pairs with valid data; {1} tsses; {2} drms", cordata.Count(), cordata.ToLookup(x => x.Tss).Count, cordata.ToLookup(x => x.Drm).Count);

                    /*Console.WriteLine("\t\t\tRunning R on data");
                    double[][] corvalues = R.TestCorrelations(cordata.Select(x => x.VectorCsv));*/

                    using (TextWriter tw = new StreamWriter(string.Format(
                        "../temp/results/MapGeneration/{3}{2}Correlation.{0}.{1}.tsv", 
                        expressionType.Key,
                        histoneType.Key, 
                        string.IsNullOrEmpty(cellLine) ? string.Empty : string.Format("{0}.", cellLine),
                        omitDrmSource ? string.Empty : "Full.")))
                    {
                        for (int i = 0; i < cordata.Count; i++)
                        {
                            allTssDrmPairs.Add(string.Format("{0}#{1}", cordata[i].Tss, cordata[i].Drm));
                            tw.WriteLine("{0}\t{1}\t{2}\t{3}", cordata[i].Tss, cordata[i].Drm, cordata[i].Pvalue, cordata[i].Correlation);
                        }
                    }

                }

                var groupedTssDrmPairs = allTssDrmPairs.ToLookup(x => x);

                Console.WriteLine("All considered TSS-DRM pairs: {0}", groupedTssDrmPairs.Count());
            }
          

            // TODO: Filter out TSSes with < 2 RPM/RPKM on all cell lines

            // TODO: Filter histone TSS by having at least one with a signal value of 2 and at least 2-fold difference between lines

            // TODO: Assign histone value and 
        }


        private static HashSet<string> FilterTranscripts(string configFile, string omittedCellLine)
        {
            XDocument doc = XDocument.Load(configFile);
            
            //string histoneRoot = doc.Descendants("HistoneRoot").First().Value.ToString();
            //string drmFileName = doc.Descendants("DrmFileName").First().Value.ToString();
            //string cellLineDrmFileName = doc.Descendants("CellLineDrmFileNames").First().Value.ToString();
            //int minFeatureCount = string.IsNullOrEmpty(omittedCellLine) ? 7 : 6;
            //BedFile drmFile = new BedFile(string.IsNullOrEmpty(omittedCellLine) ? drmFileName : string.Format(cellLineDrmFileName, omittedCellLine), BedFile.Bed4Drm);

            string expressionRoot = doc.Descendants("ExpressionRoot").First().Value.ToString();
            string[] cellLines = doc.Descendants("CellLines").First().Value.ToString().Split(',')
                .Where(cl => string.IsNullOrEmpty(omittedCellLine) || string.Compare(cl, omittedCellLine) != 0).ToArray();
            
            string[] expressionTypes = doc.Descendants("ExpressionTypes").First().Value.ToString().Split(',');

            Console.WriteLine("Building drm-tss pairs ...");
            
            // Expression data by type and cell line
            Dictionary<string, Dictionary<string, string>> validCellLineExpression = cellLines
                .Select(cl => expressionTypes
                        .Select(et => new
                        {
                    CellLine = cl,
                    ExpType = et,
                    Filename = string.Format("{0}/{1}/{2}.tsv", expressionRoot, cl, et)
                })
                .Where(expFile => File.Exists(expFile.Filename)))
                .SelectMany(x => x)
                .ToLookup(x => x.ExpType, x => x)
                .ToDictionary(x => x.Key, x => x.ToDictionary(y => y.CellLine, y => y.Filename));
            

            string testGene = "ENSG00000073111_127323094";

            var rawExpData = validCellLineExpression.Select(expressionType => 
            {
                Console.WriteLine("Loading {0} expression data ...", expressionType.Key);
                return expressionType.Value.Select(cl => 
                {
                    using (TextReader tr = new StreamReader(cl.Value))
                    {
                        Console.WriteLine("\t\tReading file {0}", cl.Value);
                        return tr.ReadToEnd()
                            .Split('\n')
                                .Where(line => !string.IsNullOrWhiteSpace(line))
                                .Select(line => 
                                        {
                                    string[] fields = line.Split('\t');
                                    return new 
                                    {
                                        CellLine = cl.Key,
                                        Tss = fields [0],
                                        Value = double.Parse(fields [1]),
                                        Chromosome = fields[2]
                                    };
                                });
                    }
                })
                .SelectMany(x => x)
                .ToLookup(x => x.Tss, x => x)
                .Where(x => x.Count() > 0)
                .Select(x => new 
                {
                    Tss = x.Key,
                    ExpressionType = expressionType.Key,
                    Chromosome = x.First().Chromosome,
                    Position = int.Parse(x.Key.Split('_')[1]),
                    Expression = x.ToLookup(y => y.CellLine, y => y.Value)
                        .ToDictionary(y => y.Key, y => y.Max())
                });
            })
            .SelectMany(x => x).ToList();

            Dictionary<string, int> tssWithMinExpAcrossAllTypes = rawExpData
                .Select(x => new {Tss = x.Tss, ExpList = x.Expression.Values })
                .Select(x => x.ExpList.Select(y => new { Tss = x.Tss, Exp = y }))
                .SelectMany(x => x)
                .ToLookup(x => x.Tss, x => x.Exp)
                .Select(x => new { Tss = x.Key, CountOver2 = x.Count(y => y >= 2) })
                .Where(x =>
                {
                        if (string.Compare(x.Tss, testGene) == 0)
                        {
                            Console.WriteLine("{1} count: {0}", x.CountOver2, testGene);
                        }
                        return x.CountOver2 > 0; 
                })
                .ToDictionary(x => x.Tss, x => x.CountOver2);

            int tssExpOver2 = tssWithMinExpAcrossAllTypes.Count();

            Console.WriteLine("Count of tss with expression over 2 across all cell lines and experiment types: {0}", tssExpOver2);

            Dictionary<string, int> tssWithMinVar = rawExpData
                .Where(x => tssWithMinExpAcrossAllTypes.ContainsKey(x.Tss))
                .Select(x => new {Tss = x.Tss, Experiment = x.ExpressionType, ExpList = x.Expression.Values })
                .Select(x => x.ExpList.Select(y => new { Tss = x.Tss, ET = x.Experiment, Exp = y }))
                .SelectMany(x => x)
                .ToLookup(x => x.Tss, x => x)
                .ToDictionary(x => x.Key, x => x.ToLookup(y => y.ET, y => y.Exp).Where(y => y.Count(z => z > 2) > 0))
                .Where(x => x.Value.Count() > 0)
                .Select(x =>
                {
                    if (string.Compare(x.Key, testGene) == 0)
                    {
                        foreach (var v in x.Value)
                        {
                                double m = v.Average();
                                Console.WriteLine(
                                    "{4} is two-fold: mean: {0}, max {1}, min {2}, var {5}: {3}", 
                                    v.Average(),
                                    v.Max(),
                                    v.Min(),
                                    IsTwoFoldChange(v),
                                    testGene,
                                    Math.Sqrt(v.Sum(a => Math.Pow(a - m, 2))));
                        }
                        
                    }
                    return new { Tss = x.Key, CountFoldChange = x.Value.Count(exptype => IsTwoFoldChange(exptype)) };
                })
                .Where(x => x.CountFoldChange > 0)
                .ToDictionary(x => x.Tss, x => x.CountFoldChange);

            int tssFolChangeOver2 = tssWithMinVar.Count();

            Console.WriteLine("{1} has count of fold change: {0}", tssWithMinVar[testGene], testGene);
            
            Console.WriteLine("Count of tss 2-fold change expression within each experiment type accross cell lines with only tss of min expression 2 somewhere: {0}", tssFolChangeOver2);



            //Console.WriteLine(string.Join("\n", tssWithMinVar.Keys));

            return new HashSet<string>(tssWithMinVar.Keys);
        }
                
        private static void GenomeWideRegression(string xmlFile, string cellLine, bool fGenerateFeatures, bool fUseVariantExpressedSubset)
        {
            Config config = new Config(xmlFile);

            string chipFileName = config.CellLineTag.Replace(config.GetChipDataSourceFileList("AllChIP"), cellLine);

            List<string> chipDataSets = GetChIPDataSetNames(chipFileName).ToList();

            int k = 0;
            if (fGenerateFeatures)
            {
                List<Dictionary<string, BedFile.Location>> tssData = GetTssDataFromExpression(
                    config.RnaSources,
                    i => config.CellLineTag.Replace(config.RnaSourceTag.Replace(config.ExpressionFileFormat, config.RnaSources [i]), cellLine));

                var universalGeneIds = UnionTssLocations(tssData);

                foreach (string filename in GetChIPDataFileNames(chipFileName))
                {
                    var chipFile = new BedFile(filename, filename.Trim().Contains("broad") ? BedFile.Bed6Plus3Layout : BedFile.Bed6Plus4Layout);

                    TRFScorer.CreateIndex(universalGeneIds, "TSS", cellLine);

                    string baseFileName = chipFile.FileName.Split('.').First();
                    Console.WriteLine("\tScoring feature {2}: {0} of {1}", k, chipDataSets.Count, baseFileName);

                    TRFScorer.WriteScoreNameValuePairs(
                        GetTssScoreFileName(baseFileName),
                        TRFScorer.CreateTssScores(universalGeneIds, chipFile, config.D0, config.Range), null);
                    
                    GC.Collect();
                    k++;
                }
                return;
            }

            Dictionary<string, int> tssIndex = TRFScorer.ReadIndex("TSS", cellLine);

            List<double[]> tssScores = new List<double[]>();
            foreach (string chipFile in chipDataSets)
            {
                Console.WriteLine("\tReading feature {2}: {0} of {1}", k++, chipDataSets.Count, GetTssScoreFileName(chipFile));
                tssScores.Add(TRFScorer.ReadScoreNameValuePairs<double>(tssIndex, GetTssScoreFileName(chipFile), x => double.Parse(x)));
                
                GC.Collect();
            }

            string[][] mapFileNames = config.RnaSources.Select(rnaSource => config.Thresholds.Select(threshhold =>
                config.CellLineTag.Replace(
                    config.RnaSourceTag.Replace(
                        config.ThresholdTag.Replace(config.MapFileFormat, threshhold), 
                        rnaSource),  
                    cellLine)).ToArray()).ToArray();

            List<Dictionary<string, int>> tssIndices;

            if (fUseVariantExpressedSubset)
            {
                List<HashSet<string>> variantExpressedTss = GetTssWithVariedExpression(cellLine, config.RnaSources, false);

                tssIndices = variantExpressedTss.Select(tssSet => tssSet
                                                        .Where(x => tssIndex.ContainsKey(x))
                                                        .Select(x => new { Tss = x, Index = tssIndex[x] })
                                                        .ToDictionary(x => x.Tss, x => x.Index)).ToList();
            }
            else
            {
                tssIndices = new List<Dictionary<string, int>>();
                for (int i = 0; i < config.SourceCount; i++)
                {
                    tssIndices.Add(tssIndex);
                }
            }


            for (int i = 0; i < config.SourceCount; i++)
            {   
                Dictionary<string, double> expression = GetExpression(config.CellLineTag.Replace(config.RnaSourceTag.Replace(config.ExpressionFileFormat, config.RnaSources[i]), cellLine));

                string tableFile = "../temp/" + config.GetGenomeWideExpressionFile(cellLine);
                int nTss = R.CreateTssScoreTable(tssScores, tssIndices[i], expression, tableFile, chipDataSets);

                List<List<string>> geneSets = new List<List<string>>();

                List<string> threshholdNames = new List<string> { "Genome Wide" };
                List<int> tssCount = new List<int> { nTss };

                for (int j = 0; j < config.Thresholds.Length; j++)
                {
                    var genes = LoadMap(mapFileNames[i][j]).Select(x => x.Key).Where(x => expression.ContainsKey(x)).ToList();

                    geneSets.Add(genes);
                    threshholdNames.Add(string.Format("TSS Log10 < -{0} Map", config.Thresholds[j]));
                    tssCount.Add(genes.Count);
                }

                double[] RsquaredMean;
                double[] RsquaredDev;
                double[] Pvals;
                double[] MapRSquareds = R.RegressDataEvaluateSubsets(tableFile, chipDataSets, R.RegressionFunction.Linear, geneSets, out RsquaredMean, out RsquaredDev, out Pvals);

                Console.WriteLine("<table>\n<tr><th colspan=6 align=center>Genome Wide Regression of {0} on {1}</th></tr>", config.RnaSources[i], cellLine);
                Console.WriteLine(trhs + "Gene subset" + th + "TSS count" + th + "Map TSS R2" + th + 
                                  "Mean GW R2" + th + "GW R2 Std Err" + th + "Map-GW P-value" + trhe);
                Console.WriteLine(string.Join("\n", MapRSquareds.Select((r, j) => 
                                  trds + threshholdNames[j] + td + tssCount[j] + td + r + td + 
                                  RsquaredMean[j] + td + (RsquaredDev[j] / Math.Sqrt(tssCount[j])) + td  + Pvals[j] + trde)));
                Console.WriteLine("</table>");
            }
        }

        private static void CrossCellLineExpressionCorrelation(string xmlFile)
        {
            Config config = new Config(xmlFile);

            string[] cellLines = config.CellLines;

            List<List<HashSet<string>>> tssWithVariedExpression = cellLines
                .Select(cellLine => GetTssWithVariedExpression(cellLine, config.RnaSources, false))
                .ToList();

            Dictionary<string, List<double>> rnaSourceCorrelationsAllVariantExpress = new Dictionary<string, List<double>>();
            Dictionary<string, Dictionary<string, List<double>>> rnaSourceCorrelationsByThreshold = new Dictionary<string, Dictionary<string, List<double>>>();

            for (int r = 0; r < config.RnaSources.Length; r++)
            {
                string rnaSource = config.RnaSources[r];

                rnaSourceCorrelationsByThreshold.Add(
                    rnaSource, 
                    config.Thresholds.ToDictionary(x => x, x => new List<double>()));

                rnaSourceCorrelationsAllVariantExpress.Add(rnaSource, new List<double>());

                for (int i = 0; i < cellLines.Length; i++)
                {
                    string baseExpressionFile = config.RnaSourceTag.Replace(config.ExpressionFileFormat, rnaSource);
                    string expressionFileA = config.CellLineTag.Replace(baseExpressionFile, cellLines[i]);

                    Console.WriteLine("Processing {0} on {1}", rnaSource, cellLines[i]);
                    var expA = GetExpression(expressionFileA)
                        .Where(x => tssWithVariedExpression[i][r].Contains(x.Key))
                        .ToDictionary(x => x.Key, x => x.Value);

                    string[][] mapFileNames = config.GetMapFileNames(cellLines[i]);

                    var otherCellLines = config.StringValues["OtherCellLines"].Split(',').ToArray();

                    for (int j = i + 1; j < otherCellLines.Length; j++)
                    {
                        if (cellLines[i] != otherCellLines[j])
                        {
                            string expressionFileB = config.CellLineTag.Replace(baseExpressionFile, otherCellLines[j]);

                            if (File.Exists(expressionFileB))
                            {
                                var expB = GetExpression(expressionFileB);

                                Dictionary<string, Stats.Point2> data = expA
                                    .Where(x => expB.ContainsKey(x.Key))
                                    .Select(x => new 
                                    {
                                        Tss = x.Key, 
                                        Point = new Stats.Point2
                                {
                                    x = x.Value,
                                    y = expB [x.Key]
                                }
                                    })
                                    .ToDictionary(x => x.Tss, x => x.Point);

                                if (data.Count() > 0)
                                {

                                    double allTssCorrelation = Stats.Pearson(data
                                        .Where(x => tssWithVariedExpression [i] [r].Contains(x.Key) &&
                                        tssWithVariedExpression [i] [r].Contains(x.Key))
                                        .Select(x => x.Value));

                                    Console.WriteLine("Correlation {0}:{1} on {2} = {3}",
                                                      config.CellLines [i],
                                                      otherCellLines [j],
                                                      rnaSource,
                                                      allTssCorrelation);

                                    rnaSourceCorrelationsAllVariantExpress [rnaSource].Add(allTssCorrelation);

                                    for (int k = 0; k < config.Thresholds.Length; k++)
                                    {
                                        var genes = LoadMap(mapFileNames [r] [k]).Select(x => x.Key).Where(x => data.ContainsKey(x)).Select(x => data [x]);

                                        double mapCorrelation = Stats.Pearson(genes);

                                        Console.WriteLine("\t{0} Threshold {1}: {2}",
                                                          rnaSource,
                                                          config.Thresholds [k],
                                                          mapCorrelation);

                                        rnaSourceCorrelationsByThreshold [rnaSource] [config.Thresholds [k]].Add(mapCorrelation);
                                    }
                                }
                            }
                        }
                    }
                }
            }

            Console.WriteLine(trhs + "Threshold</th><th colspan=2>" + string.Join("</th><th colspan=2>", rnaSourceCorrelationsAllVariantExpress.Keys) + trhe);
            Console.WriteLine(trds + "None" + td + string.Join(
                td, 
                rnaSourceCorrelationsAllVariantExpress.Select(x => x.Value.Average() + td + Stats.StdDev(x.Value))) + trde);

            Console.WriteLine(string.Join(
                "\n", 
                config.Thresholds.Select(t => trds + t + td + string.Join(
                    td, 
                    rnaSourceCorrelationsByThreshold.Select(x => x.Value[t].Average() + td + Stats.StdDev(x.Value[t]))) + trde)));
        }

        /// <summary>
        /// Write a two-column file for R regression using scoring function
        /// </summary>
        /// <param name="map">Map.</param>
        /// <param name="scorer">Scorer.</param>
        /// <param name="getExpression">Get expression.</param>
        /// <param name="cellLine">Cell line.</param>
        /// <param name="chipType">Chip type.</param>
        /// <param name="setName">Set name.</param>
        /// <param name="rnaSource">Rna source.</param>
        static string WriteTssAndTssDrmTFScores(
            IEnumerable<string> tsses,
            List<ILookup<string, string>> maps,
            List<Func<string, IEnumerable<string>, double>> scorer,
            List<string> scoreNames,
            Func<string, double> getExpression,
            string cellLine,
            string chipType,
            string rnaSource)
        {
            string filename = string.Format("../temp/results/TFGeneSets/{0}.{1}.{2}.tsv", cellLine, rnaSource, chipType);
            using (TextWriter tw = new StreamWriter(filename))
            {
                tw.WriteLine("{0}\tesc", string.Join("\t", scoreNames));
                tw.WriteLine(string.Join("\n", tsses.Select(x => string.Format(
                    "{0}\t{1}\t{2}", 
                    x, 
                    string.Join("\t", maps.Select((map, i) => map.Contains(x) ? scorer[i](x, map[x]) : 0)), 
                    getExpression(x)))));
            }

            return filename;
        }

        /// <summary>
        /// Gets tf-by-tf regression R2 z-scores (based on tss only regression) or sets of genes with tf binding in only the drm, tss, or both
        /// </summary>
        /// <param name="configFile">Config file.</param>
        /// <param name="cellLine">Cell line.</param>
        static void TFGenePairs(string configFile, string cellLine, bool getGeneSets, bool getCorrelation)
        {
            Config config = new Config(configFile);

            Console.WriteLine("Loading drms for cell line {0}", cellLine);
            BedFile drmFile = new BedFile(config.GetDrmFileNameForCellLine(cellLine), BedFile.Bed4Drm);
            
            Console.WriteLine("Loading tss data for {0}", cellLine);
            List<Dictionary<string, BedFile.Location>> tssData = GetTssDataFromExpression(
                config.RnaSources, 
                i => config.CellLineTag.Replace(config.RnaSourceTag.Replace(config.ExpressionFileFormat, config.RnaSources [i]), cellLine));

            Func<int, int, BedFile.Location, bool> Overlap = (start, end, location) => 
            {
                if ((location.Start <= start && location.End >= start) || // location crosses range start
                    (location.Start <= end && location.End >= end) || // location crosses range end
                    (location.Start >= start && location.Start <= end) || // location starts within range
                    (location.End >= start && location.End <= end))         // location ends within range
                {
                    return true;
                }
                return false;
            };

            List<ILookup<string, string>> drmMaps = config.RnaSources.Select((x, i) => LoadMap(config.GetMapFileNames(cellLine)[i][5])).ToList();

            string chipFileName = config.CellLineTag.Replace(config.GetChipDataSourceFileList("TRF"), cellLine);
            
            List<string> chipDataSets = GetChIPDataSetNames(chipFileName).ToList();

            if (getGeneSets)
            {

                List<Dictionary<string, List<BedFile.Location>>> orderedTssStarts = tssData
                    .Select(data => data
                            .GroupBy(x => x.Value.Chromosome, x => x.Value)
                            .ToDictionary(x => x.Key, x => x.OrderBy(y => y.Mid)
                            .ToList()))
                    .ToList();

                foreach (string filename in GetChIPDataFileNames(config.GetChipDataSourceFileList("TRF", cellLine)))
                {
                    Console.WriteLine("\tEvaluating peaks from {0}", filename);
                    
                    BedFile.Layout layout = filename.Trim().Contains("broad") ? BedFile.Bed6Plus3Layout : BedFile.Bed6Plus4Layout;
                    var chipFile = new BedFile(filename, layout);
                    
                    var orderedChIPPeaks = chipFile.GetOrderedLocations(x => x.Mid);

                    Dictionary<string, double> drmPeaks = drmFile.Locations
                        .Where(drm => orderedChIPPeaks.ContainsKey(drm.Value.Chromosome))
                            .Select(drm => new { Drm = drm, Peak = TRFScorer.GetClosest(orderedChIPPeaks[drm.Value.Chromosome], drm.Value.Mid) } )
                            .Where(x => Overlap(x.Drm.Value.Start, x.Drm.Value.End, x.Peak))
                            .ToDictionary(x => x.Drm.Value.Name, x => double.Parse(x.Peak.Data[layout.SignalValue]));

                    for (int i = 0; i < config.RnaSources.Count(); i++)
                    {

                        Dictionary<string, double> tssPeaks = orderedTssStarts[i]
                            .Where(chr => orderedChIPPeaks.ContainsKey(chr.Key))
                                .SelectMany(x => x.Value)
                                .Select(tss => new { Peak = TRFScorer.GetClosest(orderedChIPPeaks[tss.Chromosome], tss.Start), Tss = tss } )
                                .Where(x => Overlap(x.Tss.Start - config.Range, x.Tss.Start + config.Range, x.Peak))
                                .ToDictionary(x => x.Tss.Name, x => Math.Exp((double)Math.Abs(x.Peak.Mid - x.Tss.Start) / config.D0) * double.Parse(x.Peak.Data[layout.SignalValue]));
                        
                        var setNames = new List<string> { "DrmOnly", "TssOnly", "DrmAndTss" };
                        var setLogic = new List<Func<IGrouping<string, string>, bool>> 
                        {
                            (x) => !tssPeaks.ContainsKey(x.Key) && x.Any(y => drmPeaks.ContainsKey(y)),
                            (x) => tssPeaks.ContainsKey(x.Key) && !x.Any(y => drmPeaks.ContainsKey(y)),
                            (x) => tssPeaks.ContainsKey(x.Key) && x.Any(y => drmPeaks.ContainsKey(y))
                        };    

                        var submap = drmMaps[i]
                        .Where(x => x.Max(y => drmPeaks.ContainsKey(y) ? drmPeaks[y] : 0) > 0)
                            .Select(x => x.Select(y => new { Tss = x.Key, Drm = y }))
                                .SelectMany(x => x)
                                .ToLookup(x => x.Tss, x => x.Drm);

                        for (int j = 0; j < setLogic.Count; j++)
                        {
                            using (TextWriter tw = new StreamWriter("results/TFGeneSets/Sets/" + cellLine + "." + config.RnaSources[i] + "." 
                                                                    + filename.Split('/').Last().Split('.').First() + "." + setNames[j] + ".nsv"))
                            {
                                tw.WriteLine(string.Join("\n", submap.Where(tssDrms => setLogic[j](tssDrms)).Select(tssDrms => tssDrms.Key)));
                            }
                        }
                    }
                }
            }
            else
            {

                List<Dictionary<string, double>> ExpressionData = config.RnaSources.Select(x => GetExpression(config.GetExpressionFileName(x, cellLine))).ToList();

                List<ILookup<string, string>> filteredNNMaps = GetNNMapForVariantExpressedGenes(cellLine, config.RnaSources, drmFile, tssData, false, false);

                List<string>[] outputData = config.RnaSources.Select(x => new List<string>
                    {
                        "<tr><th>ChIP Set</th><th>Regression Weight</th><th>Map TSS Count</th><th>Map TSS R2</th><th>Map TSS-Map DRM R2</th><th>Map TSS-NN DRM R2</th>" +
                        "<th>Map TSS-Map DRM Z-score</th><th>Map TSS-NN DRM Z-score</th>" + 
                        "<th>Sampled TSS R2</th><th>Sampled TSS-NN DRM R2</th><th>Sampled TSS-NN DRM Z-score</th></tr>"
                    }).ToArray();

                Dictionary<string, int> tssIndex = TRFScorer.ReadIndex("TSS", cellLine);
                Dictionary<string, int> drmIndex = TRFScorer.ReadIndex("DRM", cellLine);
                
                List<double[]> tssOnlyScores = new List<double[]>();
                List<double[]> drmOnlyScores = new List<double[]>();
                int k = 0;
                if (!getCorrelation)
                {
                    Console.WriteLine("Reading generated features for regression");
                    foreach (string chipFile in chipDataSets)
                    {
                        Console.WriteLine("\tReading feature {2}: {0} of {1}", k++, chipDataSets.Count, GetTssScoreFileName(chipFile));
                        tssOnlyScores.Add(TRFScorer.ReadScoreNameValuePairs<double>(tssIndex, GetTssScoreFileName(chipFile), x => double.Parse(x)));
                        drmOnlyScores.Add(TRFScorer.ReadScoreNameValuePairs<double>(drmIndex, GetDrmScoreFileName(chipFile), x => double.Parse(x)));
                        
                        GC.Collect();
                    }
                }

                //Dictionary<string, double> expression = GetExpression(config.CellLineTag.Replace(config.RnaSourceTag.Replace(config.ExpressionFileFormat, config.RnaSources[i]), cellLine));

                for (int i = 0; i < config.RnaSources.Count(); i++)
                {
                    List<string> orderedChipDataSets = new List<string>(chipDataSets);
                    List<string> orderedChipFiles = GetChIPDataFileNames(config.GetChipDataSourceFileList("TRF", cellLine)).ToList();
                    Dictionary<string, double> weights = orderedChipDataSets.Select(x => new { Set = x, Value = 0.0 }).ToDictionary(x => x.Set, x => x.Value);

                    if (!getCorrelation)
                    {
                        string tableFile = "../temp/" + config.GetGenomeWideExpressionFile(cellLine + "TFOnly");
                        R.CreateTssScoreTable(tssOnlyScores, tssIndex, ExpressionData[i], tableFile, chipDataSets);
                        
                        weights = R.GetRegressWeights(tableFile, chipDataSets);
                        orderedChipDataSets = chipDataSets.OrderBy(x => -Math.Abs(weights[x])).ToList();
                        orderedChipFiles = GetChIPDataFileNames(
                            config.GetChipDataSourceFileList("TRF", cellLine)).Select((x, index) => new 
                            {
                                Data = x,
                                Index = index
                            }).OrderBy(d => -Math.Abs(weights[chipDataSets[d.Index]])).Select(d => d.Data).ToList();

                        Console.WriteLine(string.Join("\n", orderedChipDataSets.Select(x => string.Format("{0}\t{1}", x, weights[x]))));
                    }
                    else
                    {
                        Console.WriteLine("<table>" + trhs + "TF" + th + "NN Correlation" + th + "Map Correlation" + th + "P-value" + trhe);
                    }

                    int c = 0;
                    foreach (string filename in orderedChipFiles)
                    {
                        BedFile.Layout layout = filename.Trim().Contains("broad") ? BedFile.Bed6Plus3Layout : BedFile.Bed6Plus4Layout;
                        var chipFile = new BedFile(filename, layout);

                        var drmPeaks = TRFScorer.CreateDrmScores(drmFile, chipFile, true);

                        var tssScores = TRFScorer.CreateTssScores(tssData[i], chipFile, config.D0, config.Range);

                        string peakName = Regex.Split(Regex.Split(chipFile.FileName.Split('/').Last(), cellLine).Last(), "Pk").First();
                        
                        Console.WriteLine("\tEvaluating peaks from {0}", peakName);

                        if (getCorrelation)
                        {
                            var nndata = filteredNNMaps[i].Select(tss => new Stats.Point2 
                            {
                                x = Math.Log((tssScores.ContainsKey(tss.Key) ? tssScores[tss.Key] : 0.0) + 0.01), 
                                y = Math.Log((tss.Max(drm => drmPeaks.ContainsKey(drm) ? drmPeaks[drm] : 0.0) + 0.01))
                            });

                            var data = drmMaps[i].Select(tss => new Stats.Point2 
                            {
                                x = Math.Log((tssScores.ContainsKey(tss.Key) ? tssScores[tss.Key] : 0.0) + 0.01), 
                                y = Math.Log((tss.Max(drm => drmPeaks.ContainsKey(drm) ? drmPeaks[drm] : 0.0) + 0.01))
                            });

                            int repCount = 1000000;
                            var cors = Stats.Bootstrap<Stats.Point2, double>(
                                nndata.ToArray(), 
                                repCount, 
                                (x, d) => Stats.Pearson(d.Select(index => x[index])),
                                drmMaps[i].Count)
                                .Where(x => !double.IsNaN(x));
                            //Console.WriteLine(string.Join("\n", cors));
                            double mapCor = Stats.Pearson(data);
                            Console.WriteLine(trds + peakName + td + cors.Average() + td + mapCor + td + ((double)cors.Count(x => Math.Abs(x) > Math.Abs(mapCor)) / repCount) + trde);
                        }
                        else
                        {

                            var setNames = new List<string> { "TssOnly", "DrmAndTss", "NNDrmAndTss" };
                            var maps = new List<ILookup<string, string>> 
                            {
                                filteredNNMaps[i].Concat(drmMaps[i]).Select(x => x.Select(y => new { Tss = x.Key, Drm = y })).SelectMany(x => x).ToLookup(x => x.Tss, x => x.Drm), 
                                drmMaps[i], 
                                filteredNNMaps[i]
                            };

                            var scoreCombination = new List<Func<string, IEnumerable<string>, double>> 
                            {
                                (tss, drms) => tssScores.ContainsKey(tss) ? tssScores[tss] : 0.0,
                                (tss, drms) => (tssScores.ContainsKey(tss) ? tssScores[tss] : 0.0) + drms.Sum(y => drmPeaks.ContainsKey(y) ? drmPeaks[y] : 0.0),
                                (tss, drms) => (tssScores.ContainsKey(tss) ? tssScores[tss] : 0.0) + drms.Sum(y => drmPeaks.ContainsKey(y) ? drmPeaks[y] : 0.0)
                            };

                            string chipName = filename.Split('/').Last().Split('.').First();
                            string dataFile = WriteTssAndTssDrmTFScores(
                                         filteredNNMaps[i].Select(x => x.Key).Concat(drmMaps[i].Select(x => x.Key)).ToLookup(x => x).Select(x => x.Key),
                                         maps,
                                         scoreCombination,
                                         setNames,
                                         x => ExpressionData[i].ContainsKey(x) ? ExpressionData[i][x] : 0,
                                         cellLine,
                                         chipName,
                                         config.RnaSources[i]);

                            double[] rs = R.RegressGeneSetOneFeatureWithNNAndSampledControl(
                                dataFile, 
                                setNames, 
                                drmMaps[i].Select(x => x.Key).ToList());

                            double Rsquared = rs[3] == 0 || rs[4] == 0 || rs[5] == 0 ? 0 : rs[0];
                            double RsquaredDrm = rs[3] == 0 || rs[4] == 0 || rs[5] == 0 ? 0 : rs[1];

                            outputData[i].Add(string.Format(
                                "<tr><td>{0}</td><td>{1}</td><td>{2}</td><td>{3}</td><td>{4}</td><td>{5}</td><td>{6}</td><td>{7}</td><td>{8}</td><td>{9}</td><td>{10}</td></tr>",
                                chipName,
                                weights[orderedChipDataSets[c]],
                                drmMaps[i].Count(),
                                rs[3],
                                rs[4],
                                rs[5],
                                Rsquared,
                                RsquaredDrm,
                                rs[6],
                                rs[7],
                                rs[2]));
                            c++;
                            Console.WriteLine(outputData[i].Last());
                        }
                    }
                }

                if (getCorrelation)
                {
                    Console.WriteLine("</table>");
                }
                else
                {
                    for (int i = 0; i < config.RnaSources.Length; i++)
                    {
                        using (TextWriter tw = new StreamWriter("results/TFGeneSets/" + cellLine + "." + config.RnaSources[i] + ".SingleTF.html"))
                        {
                            tw.WriteLine("<table>");
                            tw.WriteLine(string.Join("\n", outputData[i]));
                            tw.WriteLine("</table>");
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Compare the DRM map and the NN map to each other for overlapping edges, drms, etc
        /// Uses only variantly expressed tsses and variant histone marked drms
        /// </summary>
        /// <param name="xmlFile">Xml file.</param>
        /// <param name="cellLine">Cell line.</param>
        public static void MapNNOverlap(string xmlFile, string cellLine)
        {
            Config config = new Config(xmlFile);

            Console.WriteLine("Loading drms for cell line {0}", cellLine);
            BedFile drmFile = new BedFile(config.GetDrmFileNameForCellLine(cellLine), BedFile.Bed4Drm);

            Console.WriteLine("Loading tss data for {0}", cellLine);
            List<Dictionary<string, BedFile.Location>> tssData = GetTssDataFromExpression(
                config.RnaSources, 
                i => config.CellLineTag.Replace(config.RnaSourceTag.Replace(config.ExpressionFileFormat, config.RnaSources [i]), cellLine));

            List<ILookup<string, string>> filteredNNMaps = GetNNMapForVariantExpressedGenes(cellLine, config.RnaSources, drmFile, tssData, false, true);

            Console.Error.WriteLine("Tss counts: " + string.Join("\t", tssData.Select(x => x.Count())));

            List<ILookup<string, string>> allDrmTssPairs = GetAllDrmsForVariantExpressed(cellLine, config.RnaSources, tssData, drmFile, false);

            Console.WriteLine("<h2>" + cellLine + "</h2>");
            for (int i = 0; i < config.RnaSources.Length; i++)
            {
                Console.WriteLine("<h3>{0}</h3>", config.RnaSources[i]);

                Console.WriteLine("<table><tr><th>Threshold</th><th>TSS Count</th>" + 
                                  "<th>Map Edges</th><th>Map DRM Count</th>" + 
                                  "<th>NN Edges</th><th>NN DRM Count</th>" + 
                                  "<th>Edge Overlap</th><th>DRM Overlap</th>" +
                                  "<th>NN Edge Fraction in Map</th><th>NN DRM Fraction in Map</th>" +
                                  //"<th>Possible Edges</th><th>Possible DRMs</th>" +
                                  //"<th>Edge Overlap p-value</th><th>DRM Overlap p-value</th>" +
                                  "<th>Median Map DRM distance</th>" + 
                                  "<th>Median NN DRM distance</th>" +
                                  "<th>Map Mean (Median) Edge Count</th>" +
                                  "<th>NN Mean (Median) Edge Count</th></tr>");
               
                List<List<int>> distanceDists = new List<List<int>>();

                for (int t = 0; t < config.Thresholds.Length; t++)
                {
                    ILookup<string, string> drmMap = LoadMap(config.GetMapFileNames(cellLine) [i] [t], tssData[i]);

                    var mapTssNotInVariantSet = drmMap.Where(x => !tssData[i].ContainsKey(x.Key)).Select(x => x.Key);;

                    Console.Error.WriteLine("Tss set from map not oin variant set: " + string.Join("\t", mapTssNotInVariantSet));

                    Dictionary<string, double> expression = GetExpression(config.CellLineTag.Replace(config.RnaSourceTag.Replace(config.ExpressionFileFormat, config.RnaSources [i]), cellLine));
                    
                    Dictionary<string, BedFile.Location> drmFilteredLocations = RemoveDrmsInPromoters(drmFile, GetOrderedTssStarts(tssData[i]))
                        .Select(chr => chr.Value.Select(drm => new { Key = drm.Name, Value = drm}))
                            .SelectMany(x => x)
                            .ToDictionary(x => x.Key, x => x.Value);
                    
                    /*HashSet<string> drmSet = fUseVariantExpressedSubset ? GetVariedElement("Drm", cellLine, config.RnaSources[i]) : new HashSet<string>(drmFilteredLocations.Keys);
                    Dictionary<string, BedFile.Location> drmLocations = drmFilteredLocations.Where(x => drmSet.Contains(x.Key)).ToDictionary(x => x.Key, x => x.Value);;
                    var orderedDrmStarts = GetOrderedFilteredLocations(drmFilteredLocations, drmSet);
                    
                    HashSet<string> tssSet = fUseVariantExpressedSubset ? GetVariedElement("Tss", cellLine, config.RnaSources[i]) : new HashSet<string>(tssData[i].Keys);
                    var orderedFilteredTssStarts = GetOrderedFilteredLocations(tssData[i], tssSet);
                    
                    var tssNNMap = CreateTssCentricNNMap(orderedFilteredTssStarts, orderedDrmStarts)
                        .Where(x => tssSet.Contains(x.Item1) && drmSet.Contains(x.Item2))
                            .ToLookup(x => x.Item1, x => x.Item2)
                            .ToDictionary(x => x.Key, x => x.ToList());*/
                   
                    var nn = filteredNNMaps [i]
                        .Where(x => drmMap.Contains(x.Key))
                        .Select(x => x.Select(y => new { Tss = x.Key, Drm = y }))
                        .SelectMany(x => x)
                        .ToLookup(x => x.Tss, x => x.Drm);

                    var dists = ReportMapOverlap(config.Thresholds[t], drmMap, nn, allDrmTssPairs[i], tssData[i], drmFile.Locations);

                    if ((config.RnaSources [i] == "Short" && config.Thresholds [t] == "2") ||
                        (config.RnaSources [i] != "Short" && config.Thresholds [t] == "10"))
                    {
                        using (TextWriter tw = new StreamWriter(cellLine + "." + config.RnaSources[i] + "." + config.Thresholds[t] + ".linkleng"))
                        {
                            tw.WriteLine(string.Join("\n", drmMap.Select(tss => tss.Select(drm => Math.Min(Math.Abs(tssData [i] [tss.Key].Mid - drmFile.Locations [drm].Start),
                                                                                                     Math.Abs(tssData [i] [tss.Key].Mid - drmFile.Locations [drm].End)))).SelectMany(x => x)));
                        }

                        using (TextWriter tw = new StreamWriter(cellLine + "." + config.RnaSources[i] + "." + config.Thresholds[t] + ".degree"))
                        {
                            tw.WriteLine(string.Join("\n", drmMap.Select(tss => tss.Count())));
                        }

                        using (TextWriter tw = new StreamWriter(cellLine + "." + config.RnaSources[i] + "." + config.Thresholds[t] + ".drmleng"))
                        {
                            tw.WriteLine(string.Join("\n", drmMap.Select(tss => tss.Select(drm => drm)).SelectMany(x => x).Distinct()
                                                     .Select(drm => Math.Abs(Math.Max(drmFile.Locations [drm].End, drmFile.Locations [drm].Start) - Math.Min(drmFile.Locations [drm].End, drmFile.Locations [drm].Start) + 1))));
                        }

                        HashSet<string> drmWithVariedMarker = GetVariedElement("Drm", cellLine, config.RnaSources[i], false);

                        using (TextWriter tw = new StreamWriter(cellLine + "." + config.RnaSources[i] + ".drmleng"))
                        {
                            var data = drmWithVariedMarker.Where(drm => !string.IsNullOrEmpty(drm));
                            tw.WriteLine(string.Join("\n", data.Select(drm => Math.Abs(Math.Max(drmFile.Locations [drm].End, drmFile.Locations [drm].Start) - Math.Min(drmFile.Locations [drm].End, drmFile.Locations [drm].Start) + 1))));
                        }

                    }

                    if (t == config.CsvIntValues["OptimalRnaSourceThresholds"][i])
                    {
                        distanceDists = dists;
                    }
                }
                Console.WriteLine("</table>");

                string[] datasets = { "map", "nn", "all" };

                /*for (int d = 0; d < datasets.Length; d++)
                {
                    using (TextWriter tw = new StreamWriter("results/MapNNOverlap/" + cellLine + "_" + datasets[d] + "_optimalDRMlengths.nsv"))
                    {
                        tw.WriteLine(string.Join("\n", distanceDists[d]));
                    }
                }*/
            }
        }

        /// <summary>
        /// Generate map and nn overlap data for a particular threshold
        /// </summary>
        /// <param name="threshold">Threshold.</param>
        /// <param name="map">Map.</param>
        /// <param name="nn">Nn.</param>
        /// <param name="all">All drm-tss pairs.</param>
        /// <param name="tssLocations">Tss locations.</param>
        /// <param name="drmLocations">Drm locations.</param>
        static List<List<int>> ReportMapOverlap(
            string threshold, 
            ILookup<string, string> map,
            ILookup<string, string> nn,
            ILookup<string, string> all,
            Dictionary<string, BedFile.Location> tssLocations, 
            Dictionary<string, BedFile.Location> drmLocations)
        {
            int mapSize = map.Count();

            var mapEdges = map.Select(x => x.Select(y => new { Tss = x.Key, Drm = y })).SelectMany(x => x);
            var nnEdges = nn.Select(x => x.Select(y => new { Tss = x.Key, Drm = y })).SelectMany(x => x);
            var allEdges = all.Select(x => x.Select(y => new { Tss = x.Key, Drm = y })).SelectMany(x => x);


            Console.Error.WriteLine("Map TSS count: " + map.Count());
            Console.Error.WriteLine("NN TSS count: " + nn.Count());

            var nnEdgeSet = new HashSet<dynamic>(nnEdges);
          
            var edgeOverlap = mapEdges.Where(x => nnEdgeSet.Contains(x));

            var distinctMapDrms = map.Select(x => x.Select(y => y)).SelectMany(x => x).ToLookup(x => x);
            var distinctNnDrms = nn.Select(x => x.Select(y => y)).SelectMany(x => x).ToLookup(x => x);
            //var distinctAllDrms = all.Select(x => x.Select(y => y)).SelectMany(x => x).ToLookup(x => x);

            var distinctDrmOverlap = distinctMapDrms.Where(x => distinctNnDrms.Contains(x.Key));

            const string td = "</td><td>";

            var mapEdgeLengths = mapEdges
                .Where(x => tssLocations.ContainsKey(x.Tss) && drmLocations.ContainsKey(x.Drm))
                .Select(x => Math.Abs(tssLocations[x.Tss].Mid - drmLocations[x.Drm].Mid)).OrderBy(x => x).ToList();

            var nnEdgeLengths = nnEdges
                .Where(x => tssLocations.ContainsKey(x.Tss) && drmLocations.ContainsKey(x.Drm))
                .Select(x => Math.Abs(tssLocations[x.Tss].Mid - drmLocations[x.Drm].Mid)).OrderBy(x => x).ToList();

            var allEdgeLengths = allEdges
                .Where(x => tssLocations.ContainsKey(x.Tss) && drmLocations.ContainsKey(x.Drm))
                .Select(x => Math.Abs(tssLocations[x.Tss].Mid - drmLocations[x.Drm].Mid)).OrderBy(x => x).ToList();

            Console.Error.WriteLine("DRM map count withing 10k: " + mapEdgeLengths.Count(x => x < 10000));
            Console.Error.WriteLine("DRM nn count withing 10k: " + nnEdgeLengths.Count(x => x < 10000));

            var mapEdgeCounts = map.Select(x => x.Count()).OrderBy(x => x).ToList();
            var nnEdgeCounts = nn.Select(x => x.Count()).OrderBy(x => x).ToList();

            Console.WriteLine("<tr><td>" + threshold + td + mapSize + td + 
                              mapEdges.Count() + td + distinctMapDrms.Count() + td + 
                              nnEdges.Count() + td + distinctNnDrms.Count() + td +
                              edgeOverlap.Count() + td + distinctDrmOverlap.Count() + td + 
                              ((double)edgeOverlap.Count() / nnEdges.Count()) + td + ((double)distinctDrmOverlap.Count() / distinctNnDrms.Count()) + td +
                              //allEdges.Count() + td + distinctAllDrms.Count + td +
                              //Stats.StatisticalTests.HypergeometricTest(allEdges.Count(), mapEdges.Count(), nnEdges.Count(), edgeOverlap.Count()) + td +
                              //Stats.StatisticalTests.HypergeometricTest(distinctAllDrms.Count(), distinctMapDrms.Count(), distinctNnDrms.Count(), distinctDrmOverlap.Count()) + td +
                              (mapEdgeLengths.Count > 0 ? mapEdgeLengths[mapEdgeLengths.Count / 2] : 0) + td +
                              (nnEdgeLengths.Count > 0 ? nnEdgeLengths[nnEdgeLengths.Count / 2] : 0) + td +
                              mapEdgeCounts.Average() + "(" + (mapEdgeCounts.Count > 0 ? mapEdgeCounts[mapEdgeCounts.Count / 2] : 0) + ")" + td +
                              nnEdgeCounts.Average() + "(" + (nnEdgeCounts.Count > 0 ? nnEdgeCounts[nnEdgeCounts.Count / 2] : 0) + ")" +"</td></tr>");

            return new List<List<int>> { mapEdgeLengths, nnEdgeLengths, allEdgeLengths };
        }

        /// <summary>
        /// Compare TFAS histone score of TSSes in the map at optimal threshold versus the variantly expressed gene sets
        /// </summary>
        /// <param name="xmlFile">Xml file.</param>
        /// <param name="cellLine">Cell line.</param>
        /// <param name="extractData">Whether to report p-values or just extract data</param>
        public static void HistoneState(string xmlFile, string cellLine, bool extractData)
        {
            Config config = new Config(xmlFile);
            
            Console.WriteLine("Loading drms for cell line {0}", cellLine);
            BedFile drmFile = new BedFile(config.GetDrmFileNameForCellLine(cellLine), BedFile.Bed4Drm);

            Console.WriteLine("Loading tss data for {0}", cellLine);
            List<Dictionary<string, BedFile.Location>> tssData = GetTssDataFromExpression(
                config.RnaSources, 
                i => config.CellLineTag.Replace(config.RnaSourceTag.Replace(config.ExpressionFileFormat, config.RnaSources [i]), cellLine));
            
            List<ILookup<string, string>> filteredNNMaps = GetNNMapForVariantExpressedGenes(cellLine, config.RnaSources, drmFile, tssData, false, false);
            
            //List<ILookup<string, string>> allDrmTssPairs = GetAllDrmsForVariantExpressed(cellLine, config.RnaSources, tssData, drmFile);

            string chipFileName = config.CellLineTag.Replace(config.GetChipDataSourceFileList("Histone"), cellLine);
            
            List<string> chipDataSets = GetChIPDataSetNames(chipFileName).ToList();
            
            Dictionary<string, int> tssIndex = TRFScorer.ReadIndex("TSS", cellLine);
            
            List<double[]> tssScores = new List<double[]>();
            int k = 0;
            foreach (string chipFile in chipDataSets)
            {
                Console.WriteLine("\tReading feature {2}: {0} of {1}", k++, chipDataSets.Count, GetTssScoreFileName(chipFile));
                tssScores.Add(TRFScorer.ReadScoreNameValuePairs<double>(tssIndex, GetTssScoreFileName(chipFile), x => double.Parse(x)));
                
                GC.Collect();
            }

            int sampleCount = 100000;
            double[] tiles = { 0.05, 0.25, 0.5, 0.75, 0.95 };

            Console.WriteLine("<h2>" + cellLine + "</h2>");
            for (int i = 0; i < config.RnaSources.Length; i++)
            {
                List<string> orderedChipDataSets = chipDataSets;
                Dictionary<string, double> weights = chipDataSets.Select(x => new { Tss = x, Data = 0.0 }).ToDictionary(x => x.Tss, x => x.Data);

                Dictionary<string, double> expression = GetExpression(config.CellLineTag.Replace(config.RnaSourceTag.Replace(config.ExpressionFileFormat, config.RnaSources[i]), cellLine));

                if (!extractData)
                {
                    string tableFile = "../temp/" + config.GetGenomeWideExpressionFile(cellLine + "HistOnly");
                    R.CreateTssScoreTable(tssScores, tssIndex, expression, tableFile, chipDataSets);

                    weights = R.GetRegressWeights(tableFile, chipDataSets);

                    orderedChipDataSets = chipDataSets.OrderBy(x => -Math.Abs(weights[x])).ToList();
                    Console.WriteLine(string.Join("\n", orderedChipDataSets.Select(x => string.Format("{0}\t{1}", x, weights[x]))));
                }

                var unorderedChipDataSetIndices = orderedChipDataSets.Select(x => chipDataSets.FindIndex(y => string.Compare(y, x) == 0)).ToList();

                int t = config.CsvIntValues["OptimalRnaSourceThresholds"][i];
                ILookup<string, string> drmMap = LoadMap(config.GetMapFileNames(cellLine) [i] [t]);

                Console.WriteLine("<h3>{0}</h3>", config.RnaSources[i]);
                
                Console.WriteLine(
                    "<table>" + trhs + "Histone" + th + "Weight" + th +
                    string.Join(th, tiles.Select(x => "Map "+x+"ile")) + th + 
                    string.Join(th, tiles.Select(x => "Sample "+x+"ile")) + th + 
                    "p-value" + trhe);

                for (int j = 0; j < orderedChipDataSets.Count; j++)
                {
                    Console.WriteLine("Processing {0}", orderedChipDataSets[j]);
                    List<double> mapScores = drmMap
                        .Where(x => tssIndex.ContainsKey(x.Key))
                        .Select(x => Math.Log10(tssScores[unorderedChipDataSetIndices[j]][tssIndex[x.Key]] + 1))
                        .OrderBy(x => x).ToList();

                    int mapSize = mapScores.Count;

                    double meanMapScore = mapScores.Average();
                    //double devMapScore = Math.Sqrt(mapScores.Average(x => Math.Pow((x - meanMapScore), 2)));

                    IEnumerable<double> tileData = tiles.Select(x => mapScores[(int)(x * mapScores.Count)]);

                    Func<double[], int[], double> Stat = (input, sample) => 
                    {
                        var sampleData = sample.Select(x => input[x]);

                        var joinedData = mapScores.Concat(sampleData).ToList();
                        var shuffling = Stats.Shuffle(joinedData.Count());

                        return Math.Abs(sampleData.Average() - meanMapScore) - Math.Abs(
                            shuffling.Where((x, index) => index % 2 == 0).Average(x => joinedData[x]) -
                            shuffling.Where((x, index) => index % 2 == 1).Average(x => joinedData[x]));
                    };

                    var filteredTssSet = filteredNNMaps[i]
                        .Where(x => tssIndex.ContainsKey(x.Key))
                        .Select(x => Math.Log10(tssScores[unorderedChipDataSetIndices[j]][tssIndex[x.Key]] + 1))
                        .ToArray();

                    List<double> genomeWideScoreSet = filteredTssSet.ToList();
                    genomeWideScoreSet.Sort();

                    double pvalue = 0.0;
                    if (!extractData)
                    {
                        double[] sampleMeans = Stats.Bootstrap(filteredTssSet, sampleCount, Stat, mapSize);
                        pvalue = (double)sampleMeans.Count(x => x < 0) / sampleCount;
                    }
                    else
                    {
                        using (TextWriter tw = new StreamWriter("../temp/results/HistoneState/TssData." + cellLine + "." + orderedChipDataSets[j] + "." + config.RnaSources[i] + ".nsv"))
                        {
                            tw.WriteLine(string.Join("\n", drmMap
                                   .Where(x => tssIndex.ContainsKey(x.Key) && expression.ContainsKey(x.Key))
                                   .Select(x => string.Format("{0}\t{1}",
                                                       tssScores[unorderedChipDataSetIndices[j]][tssIndex[x.Key]],
                                                       expression[x.Key]))));
                        }
                        using (TextWriter tw = new StreamWriter("../temp/results/HistoneState/VEData." + cellLine +  "." + orderedChipDataSets[j] + "." + config.RnaSources[i] + ".nsv"))
                        {
                            tw.WriteLine(string.Join("\n", filteredNNMaps[i]
                                .Where(x => tssIndex.ContainsKey(x.Key) && expression.ContainsKey(x.Key))
                                .Select(x => string.Format("{0}\t{1}",
                                                       tssScores[unorderedChipDataSetIndices[j]][tssIndex[x.Key]],
                                                       expression[x.Key]))));
                        }
                    }

                    IEnumerable<double> sampleTileData = tiles.Select(x => genomeWideScoreSet[(int)(x * genomeWideScoreSet.Count)]);

                    Console.WriteLine(trds + orderedChipDataSets[j] + td + weights[orderedChipDataSets[j]] + td + 
                                      string.Join(td, tileData) + td + 
                                      string.Join(td, sampleTileData) + td + 
                                      pvalue + trde);
                }

                Console.WriteLine("</table>");
            }
        }

        /// <summary>
        /// Creates normalized binned counts of histone peaks in 10kb region around tsses for 
        /// map at optimal threshold and control of all other variantly expressed tsses
        /// </summary>
        /// <param name="xmlFile">Xml file.</param>
        /// <param name="cellLine">Cell line.</param>
        public static void HistoneProfile(string xmlFile, string cellLine)
        {
            Config config = new Config(xmlFile);

            Console.WriteLine("Loading tss data for {0}", cellLine);
            List<Dictionary<string, BedFile.Location>> tssData = GetTssDataFromExpression(
                config.RnaSources, 
                i => config.CellLineTag.Replace(config.RnaSourceTag.Replace(config.ExpressionFileFormat, config.RnaSources [i]), cellLine));

            string chipFileName = config.CellLineTag.Replace(config.GetChipDataSourceFileList("Histone"), cellLine);
            
            var chipDataSets = GetChIPDataFileNames(chipFileName);

            List<HashSet<string>> tssWithVariedExpression = GetTssWithVariedExpression(cellLine, config.RnaSources, false);

            //HashSet<string> allTssesWithAnyVariedExpression = new HashSet<string>(tssWithVariedExpression.SelectMany(x => x));

            //List<Dictionary<string, ILookup<string, BedFile.Location>>> tssPeakData = new List<Dictionary<string, ILookup<string, BedFile.Location>>>();

            for (int i = 0; i < config.RnaSources.Length; i++)
            {
                int t = config.CsvIntValues["OptimalRnaSourceThresholds"][i];
                ILookup<string, string> drmMap = LoadMap(config.GetMapFileNames(cellLine) [i] [t]);
                
                Console.WriteLine("<h2><b>" + config.RnaSources[i] + "</b></h2>");
                
                int windowHalfSize = 1000;
                int promoterHalfSize = 10000;

                foreach (string file in chipDataSets)
                {
                    var chipFile = new BedFile(file, file.Trim().Contains("broad") ? BedFile.Bed6Plus3Layout : BedFile.Bed6Plus4Layout);
                    string baseFileName = Regex.Split(Regex.Split(chipFile.FileName.Split('.').First().Split('/').Last(), cellLine).Last(), "Std").First();

                    var chipSet = new 
                    {
                        ChipName = baseFileName,
                        ChipFile = chipFile
                    };
                
                    var tssPeaks = tssWithVariedExpression[i]
                        .Where(tss => tssData[i].ContainsKey(tss))
                        .Select(tss => tssData[i][tss])
                        .Where(tss => chipSet.ChipFile.ChromosomeOrderedLocations.ContainsKey(tss.Chromosome))
                        .Select(tss => 
                            TRFScorer.EvaluateRange(
                            chipSet.ChipFile.ChromosomeOrderedLocations [tss.Chromosome], 
                            tss.Start - (promoterHalfSize + windowHalfSize), 
                            tss.End + (promoterHalfSize + windowHalfSize), 
                            new List<BedFile.Location>(),
                            (item, start) => item.Mid < start,
                            (item, end) => item.Mid <= end,
                            (items, length, location) => 
                            {
                                items.Add(location);
                                return items; 
                            })
                            .Select(y => 
                            {
                                int offset = tss.Strand == "+" ? y.Mid - tss.Mid : tss.Mid - y.Mid;
                                return new { Tss = tss.Name, Peak = offset };
                            })
                        ).SelectMany(x => x)
                        .ToLookup(x => x.Tss, x => x.Peak);

                    var offsets = drmMap.Select(x => tssPeaks[x.Key]).SelectMany(x => x);
                    var controlOffsets = tssWithVariedExpression[i].Where(tss => !drmMap.Contains(tss)).Select(x => tssPeaks[x]).SelectMany(x => x);

                    var windowCenters = Stats.Sequence(-promoterHalfSize,promoterHalfSize, 1);
                    Console.WriteLine("Window Centers: {0}", windowCenters.Count());

                    Func<IEnumerable<int>, double[]> Smoothe = offsetData =>
                    {
                        int totalPeaks = offsetData.Count();
                        var offsetCounts = offsetData.ToLookup(x => x).ToDictionary(x => x.Key, x => (double)x.Count() / totalPeaks);

                        //Console.WriteLine(string.Join("\n", offsetCounts.OrderBy(x => x.Key).Select(x => string.Format("{0}\t{1}", x.Key, x.Value))));

                        Console.WriteLine("Count sum = {0}", offsetCounts.Sum(x => x.Value));
                        double[] profile = new double[windowCenters.Count];

                        //double totalWeight = 2 * windowHalfSize * positionWeight;

                        double peakSum = 0;

                        /*for (int k = -windowHalfSize; k < windowHalfSize; k++)
                        {
                            int wc = windowCenters[0] + k;
                            peakSum += (offsetCounts.ContainsKey(wc) ? offsetCounts[wc] : 0.0);
                        }

                        for (int k = 0; k < windowCenters.Count; k++)
                        {
                            int wc = windowCenters[k] + windowHalfSize;
                            peakSum += (offsetCounts.ContainsKey(wc) ? offsetCounts[wc] : 0.0);

                            profile[k] = peakSum;

                            wc = windowCenters[k] - windowHalfSize ;
                            peakSum -= (offsetCounts.ContainsKey(wc) ? offsetCounts[wc] : 0.0);

                        }*/

                        for (int j = 0; j < windowCenters.Count; j++)
                        {
                            int wc = windowCenters[j];
                            peakSum = 0;
                            for (int k = wc - windowHalfSize; k <= wc + windowHalfSize; k++)
                            {
                                peakSum += Math.Exp(-(double)Math.Abs(k - wc) / (windowHalfSize / 2)) * (offsetCounts.ContainsKey(k) ? offsetCounts[k] : 0.0);
                                    //(1 - Math.Pow(Math.Abs(k - wc), 2.0) / Math.Pow(windowHalfSize, 2.0)) * (offsetCounts.ContainsKey(k) ? offsetCounts[k] : 0.0);
                            }

                            profile[j] = peakSum;
                        }

                        double smoothedSum = profile.Sum(x => x / windowCenters.Count);
                        Console.WriteLine("SmoothedSum: {0}", smoothedSum);

                        var ret = profile.Select(x => x / smoothedSum / windowCenters.Count * (2 * windowHalfSize)).ToArray();
                        Console.WriteLine("NormSmSum: {0}", ret.Sum());
                        return ret;
                    };
                                                  
                    var peakProfile = Smoothe(offsets);
                    var controlPeakProfile = Smoothe(controlOffsets);
                    //var controlProfile = windowCenters.Select((pos, index) => controlOffsets.Where(x => x >= pos - 100 && x <= pos + 100).Count() * windowControlWeights[index]).ToList();
                        //controlOffsets.Select(x => x / 250).ToLookup(x => x).ToDictionary(x => x.Key, x => (double)(x.Count()) / totalControlPeaks);
                    Console.WriteLine("<h3>{0}:{1} Profile:</h3>", chipSet.ChipName, config.RnaSources[i]);
                    Console.WriteLine("<table>" + trhs + "Offset" + th + "Control Peaks" + th + "Map Peaks" + trhe);
                    Console.WriteLine(string.Join("\n", windowCenters
                                      .Select((x, index) => new { X = x, I = index})
                                      //.Where(x => x.X % (2 * windowHalfSize) == 0 /*&& (x.I < 100 || (x.I > 9900 && x.I < 10100))*/)
                                      .Select(x => trds + x.X + td + controlPeakProfile[x.I] + td + peakProfile[x.I] + trde)));
                    Console.WriteLine("</table>");

                    string filename = Path.GetTempFileName();
                    using (TextWriter tw = new StreamWriter(filename))
                    {
                        tw.WriteLine(string.Join("\n", windowCenters
                                                 .Select((x, index) => x + "\t" + controlPeakProfile[index] + "\t" + peakProfile[index])));
                    }

                    R.PlotHistoneProfile(filename, cellLine, config.RnaSources[i], chipSet.ChipName);

                    Console.WriteLine(filename);
                }

            }
        
        }

        /// <summary>
        /// Perform genome wide regression (or on variantly expressed subset) on progressively more complicated models
        /// by adding TSS-centric NN, DRM-centric NN and K NN TSS-centric 
        /// </summary>
        /// <param name="xmlFile">Xml file.</param>
        /// <param name="cellLine">Cell line.</param>
        public static void ProgressivePromoterAdditions(string xmlFile, string cellLine, bool fUseVariantExpressedSubset)
        {
            Config config = new Config(xmlFile);
            
            string chipFileName = config.CellLineTag.Replace(config.GetChipDataSourceFileList("AllChIP"), cellLine);
            
            List<string> chipDataSets = GetChIPDataSetNames(chipFileName).ToList();


            var drmIndex = TRFScorer.ReadIndex("DRM", cellLine);
            var tssIndex = TRFScorer.ReadIndex("TSS", cellLine);
            
            List<double[]> drmScores = new List<double[]>();
            List<double[]> tssScores = new List<double[]>();

            int k = 0;
            foreach (string chipFile in chipDataSets)
            {
                Console.WriteLine("\tReading feature {2}: {0} of {1}", k++, chipDataSets.Count, GetDrmScoreFileName(chipFile));
                drmScores.Add(TRFScorer.ReadScoreNameValuePairs<double>(drmIndex, GetDrmScoreFileName(chipFile), x => double.Parse(x)));
                tssScores.Add(TRFScorer.ReadScoreNameValuePairs<double>(tssIndex, GetTssScoreFileName(chipFile), x => double.Parse(x)));
                
                GC.Collect();
            }

            Console.WriteLine("Loading drms for cell line {0}", cellLine);
            BedFile drmFile = new BedFile(config.GetDrmFileNameForCellLine(cellLine), BedFile.Bed4Drm);

            Console.WriteLine("Loading tss data for {0}", cellLine);
            List<Dictionary<string, BedFile.Location>> tssData = GetTssDataFromExpression(
                config.RnaSources, 
                i => config.CellLineTag.Replace(config.RnaSourceTag.Replace(config.ExpressionFileFormat, config.RnaSources[i]), cellLine));

            for (int i = 0; i < config.RnaSources.Length; i++)
            {   
                Dictionary<string, double> expression = GetExpression(config.CellLineTag.Replace(config.RnaSourceTag.Replace(config.ExpressionFileFormat, config.RnaSources [i]), cellLine));

                Dictionary<string, BedFile.Location> drmFilteredLocations = RemoveDrmsInPromoters(drmFile, GetOrderedTssStarts(tssData[i]))
                    .Select(chr => chr.Value.Select(drm => new { Key = drm.Name, Value = drm}))
                    .SelectMany(x => x)
                    .ToDictionary(x => x.Key, x => x.Value);

                HashSet<string> drmSet = fUseVariantExpressedSubset ? GetVariedElement("Drm", cellLine, config.RnaSources[i], false) : new HashSet<string>(drmFilteredLocations.Keys);
                Dictionary<string, BedFile.Location> drmLocations = drmFilteredLocations.Where(x => drmSet.Contains(x.Key)).ToDictionary(x => x.Key, x => x.Value);;
                var orderedDrmStarts = GetOrderedFilteredLocations(drmFilteredLocations, drmSet);

                HashSet<string> tssSet = fUseVariantExpressedSubset ? GetVariedElement("Tss", cellLine, config.RnaSources[i], false) : new HashSet<string>(tssData[i].Keys);
                var orderedFilteredTssStarts = GetOrderedFilteredLocations(tssData[i], tssSet);

                var tssNNMap = CreateTssCentricNNMap(orderedFilteredTssStarts, orderedDrmStarts)
                    .Where(x => tssSet.Contains(x.Item1) && drmSet.Contains(x.Item2))
                    .ToLookup(x => x.Item1, x => x.Item2)
                    .ToDictionary(x => x.Key, x => x.ToList());

                var drmNNMap = CreateDrmCentricNNMap(orderedFilteredTssStarts, orderedDrmStarts, drmLocations)
                    .Where(x => tssSet.Contains(x.Item1) && drmSet.Contains(x.Item2))
                    .ToLookup(x => x.Item1, x => x.Item2)
                    .ToDictionary(x => x.Key, x => x.ToList());

                ILookup<string, int> ExpressionTSSLocations = GetTSSMap(expression.Keys);

                Dictionary<string, string> ChIPToExpressionTSS = NearestTSS(tssNNMap.Select(x => x.Key), ExpressionTSSLocations);

                Console.WriteLine("TSS NN Map has {0} TSSes", tssNNMap.Count());
                Console.WriteLine("DRM NN Map has {0} TSSes", drmNNMap.Count());

                string tableFile = "../temp/PPA." + cellLine + ".tsv";
                int nTss = R.CreateTable(tssNNMap, tssIndex, tssScores, drmIndex, drmScores, expression, ChIPToExpressionTSS, tableFile, chipDataSets);

                var tssColumns = R.MakeTssColumnNames(chipDataSets);
                var drmColumns = R.MakeDrmColumnNames(chipDataSets);

                int repCount = 1000;

                double[] allTss_PromOnly_RSquareds = R.RegressColumns("../temp/PPA/allTss_PO.R", tableFile, tssColumns, R.RegressColumnsParam.AllRows, repCount, 1);
                double[] allTss_TssNN_RSquareds = R.RegressColumns("../temp/PPA/allTss_NN.R", tableFile, tssColumns.Concat(drmColumns), R.RegressColumnsParam.AllRows, repCount, 1);

                Console.WriteLine("All tss count: " + nTss);
                Console.WriteLine("All tss prom only: {0}", allTss_PromOnly_RSquareds.Average());
                Console.WriteLine("All tss tss NN: {0}", allTss_TssNN_RSquareds.Average());

                var tssNNMapDrmNNSubset = tssNNMap.Where(x => drmNNMap.ContainsKey(x.Key)).ToDictionary(x => x.Key, x => x.Value);

                string dataFileDrmNN = "../temp/PPA2." + cellLine + ".tsv";
                int drmNNTssCount = R.CreateTable(drmNNMap, tssIndex, tssScores, drmIndex, drmScores, expression, ChIPToExpressionTSS, dataFileDrmNN, chipDataSets);

                string dataFileTssNNDrmSubset = "../temp/PPA3." + cellLine + ".tsv";
                R.CreateTable(tssNNMapDrmNNSubset, tssIndex, tssScores, drmIndex, drmScores, expression, ChIPToExpressionTSS, dataFileTssNNDrmSubset, chipDataSets);

                double[] subsetTss_PromOnly_RSquareds = R.RegressColumns("../temp/PPA/subsetTss_PO.R", dataFileDrmNN, tssColumns, R.RegressColumnsParam.AllRows, repCount, 1);
                double[] subsetTss_DrmNN_RSquareds = R.RegressColumns("../temp/PPA/subsetTss_DRMNN.R", dataFileDrmNN, tssColumns.Concat(drmColumns), R.RegressColumnsParam.AllRows, repCount, 1);
                double[] subsetTss_TssNN_RSquareds = R.RegressColumns("../temp/PPA/subsetTss_TSSNN.R", dataFileTssNNDrmSubset, tssColumns.Concat(drmColumns), R.RegressColumnsParam.AllRows, repCount, 1);



                //double pvalue;
                //double[] subsetTss_TssNN_ShuffledDRMs = R.RegressPromAndShuffledDrm(dataFile, chipDataSets, out pvalue);

                //Console.WriteLine("Shuffled DRMs R2s: prom: " + subsetTss_TssNN_ShuffledDRMs[0] + "; tss shuffled: " + subsetTss_TssNN_ShuffledDRMs[1] + "; pvalue = " + pvalue);

                Console.WriteLine("Subset tss count: " + drmNNTssCount);
                Console.WriteLine("Subset tss prom only: {0}", subsetTss_PromOnly_RSquareds.Average());
                Console.WriteLine("Subset tss drm NN: {0}", subsetTss_DrmNN_RSquareds.Average());
                Console.WriteLine("Subset tss tss NN: {0}", subsetTss_TssNN_RSquareds.Average());

                Console.WriteLine("<table>" + trhs + "TSS Set" + th + "TSS Count" + trhe);
                Console.WriteLine(trds + "All TSS" + (fUseVariantExpressedSubset ? " Variantly Expressed" : "") + td + nTss + trde);
                Console.WriteLine(trds + "DRM NN TSS" + (fUseVariantExpressedSubset ? " Variantly Expressed" : "") + td + drmNNTssCount + trde + "</table>");

                Console.WriteLine("<table>" + trhs + "Regression Model" + th + "Adjusted R2" + trhe);
                Console.WriteLine(trds + "All TSS" + (fUseVariantExpressedSubset ? " Variantly Expressed" : " ") + "Promoter Only" + td + allTss_PromOnly_RSquareds.Average() + trde);
                Console.WriteLine(trds + "All TSS" + (fUseVariantExpressedSubset ? " Variantly Expressed" : " ") + "TSS NN Map" + td + allTss_TssNN_RSquareds.Average() + trde);
                Console.WriteLine(trds + "DRM NN TSS" + (fUseVariantExpressedSubset ? " Variantly Expressed" : " ") + "Promoter Only" + td + subsetTss_PromOnly_RSquareds.Average() + trde);
                Console.WriteLine(trds + "DRM NN TSS" + (fUseVariantExpressedSubset ? " Variantly Expressed" : " ") + "DRM NN Map" + td + subsetTss_DrmNN_RSquareds.Average() + trde);
                Console.WriteLine(trds + "DRM NN TSS" + (fUseVariantExpressedSubset ? " Variantly Expressed" : " ") + "TSS NN Map" + td + subsetTss_TssNN_RSquareds.Average() + trde + "</table>");

                Console.WriteLine("<table>" + trhs + "Comparison" + th + "pvalue" + trhe);
                Console.WriteLine(trds + "All TSS Prom vs TSS NN Map" + td + Stats.StatisticalTests.PermutationTest(allTss_PromOnly_RSquareds, allTss_TssNN_RSquareds, 100000) + trde);
                Console.WriteLine(trds + "DRM NN TSS Prom vs DRM NN Map" + td + Stats.StatisticalTests.PermutationTest(subsetTss_PromOnly_RSquareds, subsetTss_DrmNN_RSquareds, 100000) + trde);
                Console.WriteLine(trds + "DRM NN TSS Prom vs TSS NN Map" + td + Stats.StatisticalTests.PermutationTest(subsetTss_PromOnly_RSquareds, subsetTss_TssNN_RSquareds, 100000) + trde);
                Console.WriteLine(trds + "DRM NN TSS DRM NN Map vs TSS NN Map" + td + Stats.StatisticalTests.PermutationTest(subsetTss_DrmNN_RSquareds, subsetTss_TssNN_RSquareds, 100000) + trde);
            }
        }
    }
}
