using System;
using System.Collections.Generic;
using System.Linq;
using System.IO;
using System.Text.RegularExpressions;

namespace EnhancerEvaluation
{
    public class R
    {
        public R()
        {
        }

        private static readonly string ScriptPath = "source/EnhancerEvaluation/EnhancerEvaluation/RegressionHelpers.R";

        public static int CreateTable(
            Dictionary<string, List<string>> drmMap, 
            List<Dictionary<string, double>> tssScores, 
            List<Dictionary<string, double>> drmScores, 
            Dictionary<string, double> expression,
            Dictionary<string, string> ChIPToExpressionTSS,
            string outFileName, 
            List<string> chipSets)
        {
            var v = drmMap
                .Where(mapEntry => ChIPToExpressionTSS.ContainsKey(mapEntry.Key) && expression.ContainsKey(ChIPToExpressionTSS[mapEntry.Key]))
                    .Select(mapEntry => new
                    {
                        Tss = mapEntry.Key,
                        TssOnlyScore = tssScores.Select(chipData => chipData.ContainsKey(mapEntry.Key) ? chipData[mapEntry.Key] : 0.0),
                        Score = drmScores.Select(chipData => mapEntry.Value.Where(drm => chipData.ContainsKey(drm)).Sum(drm => chipData[drm])),
                        Exp = ChIPToExpressionTSS.ContainsKey(mapEntry.Key) ? 
                            ((expression.ContainsKey(ChIPToExpressionTSS[mapEntry.Key]) ?
                              expression[ChIPToExpressionTSS[mapEntry.Key]] : 0.0)) :
                                0.0
                    });
            
            Console.WriteLine("Number of TSSes: {0}", v.Count());
            
            using (TextWriter tw = new StreamWriter(outFileName))
            {
                // table format
                tw.WriteLine("{0}\t{1}\tesc", 
                             string.Join("\t", chipSets.Select(x => string.Format("Tss_{0}", x.Trim().Split('/').Last().Split('.').First().Split('-')[0]))), 
                             string.Join("\t", chipSets.Select(x => string.Format("Drm_{0}", x.Trim().Split('/').Last().Split('.').First().Split('-')[0]))));
                foreach (var gene in v)
                {
                    tw.WriteLine("{0}\t{1}\t{2}\t{3}", gene.Tss, string.Join("\t", gene.TssOnlyScore), string.Join("\t", gene.Score), gene.Exp);
                }
            }
            
            return v.Count();
        }

        /// <summary>
        /// Creates table of TSS and DRM data for regression using flat indexed tss and drm scores
        /// </summary>
        /// <returns>The table.</returns>
        /// <param name="drmMap">Drm map.</param>
        /// <param name="tssIndex">Tss index.</param>
        /// <param name="tssScores">Tss scores.</param>
        /// <param name="drmIndex">Drm index.</param>
        /// <param name="drmScores">Drm scores.</param>
        /// <param name="expression">Expression.</param>
        /// <param name="ChIPToExpressionTSS">Ch IP to expression TS.</param>
        /// <param name="outFileName">Out file name.</param>
        /// <param name="chipSets">Chip sets.</param>
        public static int CreateTable(
            Dictionary<string, List<string>> drmMap, 
            Dictionary<string, int> tssIndex,
            List<double[]> tssScores, 
            Dictionary<string, int> drmIndex,
            List<double[]> drmScores, 
            Dictionary<string, double> expression,
            Dictionary<string, string> ChIPToExpressionTSS,
            string outFileName, 
            List<string> chipSets)
        {
            var v = drmMap
                .Where(mapEntry => ChIPToExpressionTSS.ContainsKey(mapEntry.Key) && 
                       expression.ContainsKey(ChIPToExpressionTSS[mapEntry.Key]) &&
                       tssIndex.ContainsKey(mapEntry.Key))
                    .Select(mapEntry => new
                     {
                        Tss = mapEntry.Key,
                        TssOnlyScore = tssScores.Select(chipData => chipData[tssIndex[mapEntry.Key]]),
                        Score = drmScores.Select(chipData => mapEntry.Value.Where(drm => drmIndex.ContainsKey(drm)).Sum(drm => chipData[drmIndex[drm]])),
                        Exp = ChIPToExpressionTSS.ContainsKey(mapEntry.Key) ? 
                            ((expression.ContainsKey(ChIPToExpressionTSS[mapEntry.Key]) ?
                              expression[ChIPToExpressionTSS[mapEntry.Key]] : 0.0)) :
                                0.0
                    });
            
            Console.WriteLine("Number of TSSes: {0}", v.Count());
            
            using (TextWriter tw = new StreamWriter(outFileName))
            {
                // table format
                tw.WriteLine("{0}\t{1}\tesc", 
                             string.Join("\t", chipSets.Select(x => string.Format("Tss_{0}", x.Trim().Split('/').Last().Split('.').First().Split('-')[0]))), 
                             string.Join("\t", chipSets.Select(x => string.Format("Drm_{0}", x.Trim().Split('/').Last().Split('.').First().Split('-')[0]))));
                foreach (var gene in v)
                {
                    tw.WriteLine("{0}\t{1}\t{2}\t{3}", gene.Tss, string.Join("\t", gene.TssOnlyScore), string.Join("\t", gene.Score), gene.Exp);
                }
            }
            
            return v.Count();
        }

        /// <summary>
        /// Creates table of TSS and DRM data for regression using flat indexed tss and drm scores
        /// </summary>
        /// <returns>The table.</returns>
        /// <param name="drmMap">Drm map.</param>
        /// <param name="tssIndex">Tss index.</param>
        /// <param name="tssScores">Tss scores.</param>
        /// <param name="drmIndex">Drm index.</param>
        /// <param name="drmScores">Drm scores.</param>
        /// <param name="expression">Expression.</param>
        /// <param name="ChIPToExpressionTSS">Ch IP to expression TS.</param>
        /// <param name="outFileName">Out file name.</param>
        /// <param name="chipSets">Chip sets.</param>
        public static int CreateTableShuffle(
            Dictionary<string, List<string>> drmMap, 
            Dictionary<string, int> tssIndex,
            List<double[]> tssScores, 
            Dictionary<string, int> drmIndex,
            List<double[]> drmScores, 
            Dictionary<string, double> expression,
            Dictionary<string, string> ChIPToExpressionTSS,
            string outFileName, 
            List<string> chipSets)
        {
            var v = drmMap
                .Where(mapEntry => ChIPToExpressionTSS.ContainsKey(mapEntry.Key) && expression.ContainsKey(ChIPToExpressionTSS[mapEntry.Key]))
                    .Select(mapEntry => new
                            {
                        Tss = mapEntry.Key,
                        TssOnlyScore = tssScores.Select(chipData => chipData[tssIndex[mapEntry.Key]]),
                        Score = drmScores.Select(chipData => mapEntry.Value.Where(drm => drmIndex.ContainsKey(drm)).Sum(drm => chipData[drmIndex[drm]])),
                        Exp = ChIPToExpressionTSS.ContainsKey(mapEntry.Key) ? 
                            ((expression.ContainsKey(ChIPToExpressionTSS[mapEntry.Key]) ?
                              expression[ChIPToExpressionTSS[mapEntry.Key]] : 0.0)) :
                                0.0
                    });
            
            Console.WriteLine("Number of TSSes: {0}", v.Count());
            
            using (TextWriter tw = new StreamWriter(outFileName))
            {
                // table format
                tw.WriteLine("{0}\t{1}\tesc", 
                             string.Join("\t", chipSets.Select(x => string.Format("Tss_{0}", x.Trim().Split('/').Last().Split('.').First().Split('-')[0]))), 
                             string.Join("\t", chipSets.Select(x => string.Format("Drm_{0}", x.Trim().Split('/').Last().Split('.').First().Split('-')[0]))));

                //int[] sinds = Stats.Shuffle
                foreach (var gene in v)
                {
                    tw.WriteLine("{0}\t{1}\t{2}\t{3}", gene.Tss, string.Join("\t", gene.TssOnlyScore), string.Join("\t", gene.Score), gene.Exp);
                }
            }
            
            return v.Count();
        }

        private static int batchSize = 100;

        /// <summary>
        /// Creates regression table with list of values for all DRMs within megabase of tss. Count column is provided for sampling guidance of drms. 
        /// Drm scores are to be overwritten by sampling
        /// </summary>
        /// <param name="drmMap">Drm map.</param>
        /// <param name="drmMapTssAllDrms">Drm map tss all drms.</param>
        /// <param name="tssIndex">Tss index.</param>
        /// <param name="tssScores">Tss scores.</param>
        /// <param name="drmIndex">Drm index.</param>
        /// <param name="drmScores">Drm scores.</param>
        /// <param name="expression">Expression.</param>
        /// <param name="ChIPToExpressionTSS">Ch IP to expression TS.</param>
        /// <param name="outFileName">Out file name.</param>
        /// <param name="chipSets">Chip sets.</param>
        public static List<string> CreateTableAndList(
            Dictionary<string, List<string>> drmMap, 
            Dictionary<string, List<string>> drmMapTssAllDrms,
            Dictionary<string, int> tssIndex,
            List<double[]> tssScores, 
            Dictionary<string, int> drmIndex,
            List<double[]> drmScores, 
            Dictionary<string, double> expression,
            Dictionary<string, string> ChIPToExpressionTSS,
            string outFileName, 
            List<string> chipSets,
            int repCount)
        {
            var v = drmMap
                .Where(mapEntry => ChIPToExpressionTSS.ContainsKey(mapEntry.Key) && expression.ContainsKey(ChIPToExpressionTSS [mapEntry.Key]))
                    .Select(mapEntry => new
                            {
                        Tss = mapEntry.Key,
                        TssOnlyScore = tssScores.Select(chipData => chipData [tssIndex [mapEntry.Key]]),
                        Count = mapEntry.Value.Where(drm => drmIndex.ContainsKey(drm)).Count(),
                        Score = drmScores.Select(chipData => mapEntry.Value.Where(drm => drmIndex.ContainsKey(drm)).Sum(drm => chipData[drmIndex[drm]])),
                        Exp = ChIPToExpressionTSS.ContainsKey(mapEntry.Key) ? 
                            ((expression.ContainsKey(ChIPToExpressionTSS [mapEntry.Key]) ?
                              expression [ChIPToExpressionTSS [mapEntry.Key]] : 0.0)) :
                                0.0
                    });

            Console.WriteLine("Number of TSSes: {0}", v.Count());
            
            using (TextWriter tw = new StreamWriter(outFileName))
            {
                // table format
                tw.WriteLine(
                    string.Join("\t", chipSets.Select(x => string.Format("Tss_{0}", x.Trim().Split('/').Last().Split('.').First().Split('-') [0]))) + "\t" +
                    string.Join("\t", chipSets.Select(x => string.Format("Drm_{0}", x.Trim().Split('/').Last().Split('.').First().Split('-')[0]))) + 
                    "\tesc");

                foreach (var gene in v)
                {
                    tw.WriteLine(gene.Tss + "\t" +
                                 string.Join("\t", gene.TssOnlyScore) + "\t" +
                                 string.Join("\t", gene.Score) + "\t" +
                                 gene.Exp);
                }
            }

            List<int> batchSequence = Stats.Sequence(0, repCount / batchSize - 1, 1);
            List<string> batchFiles = new List<string>();

            foreach (int i in batchSequence)
            {
                string batchFile = outFileName.Replace(".tsv", "." + i + ".tsv");


                using (TextWriter tw = new StreamWriter(batchFile))
                {
                    // table format
                    tw.WriteLine(
                        string.Join("\t", chipSets.Select(x => string.Format("Tss_{0}", x.Trim().Split('/').Last().Split('.').First().Split('-') [0]))) + "\t" +
                        string.Join("\t", chipSets.Select(x => string.Format("Drm_{0}", x.Trim().Split('/').Last().Split('.').First().Split('-')[0]))) + 
                        "\tesc");

                    for (int j = 0; j < batchSize; j++)
                    {
                        foreach (var gene in v)
                        {
                            var sampledDrmScores = gene.Score.Select((x, index) => Stats.Sample(gene.Count, drmMapTssAllDrms[gene.Tss].Count)
                                                                     .Sum(sampleIndex => drmScores[index][drmIndex[drmMapTssAllDrms[gene.Tss][sampleIndex]]]));
                            tw.WriteLine( gene.Tss + "_" + j + "\t" +
                                         string.Join("\t", gene.TssOnlyScore) + "\t" +
                                         string.Join("\t", sampledDrmScores) + "\t" +
                                         gene.Exp);
                        }
                    }
                }

                batchFiles.Add(batchFile);
            }

            //string listScriptFile = outFileName + ".R";
            /*using (TextWriter tw = new StreamWriter(listScriptFile))
            {
                tw.WriteLine("drmdata <- function() {");

                for (int i = 0; i < drmScores.Count; i++)
                {
                    foreach (var gene in v)
                    {
                        tw.WriteLine("list" + gene.Tss + "_" + i +" <- c(" +
                                     string.Join(", ", drmMapTssAllDrms.ContainsKey(gene.Tss) ? drmMapTssAllDrms[gene.Tss].Select(drm => drmScores[i][drmIndex[drm]]) : new List<double>(0)) + ")");
                    }
                    // Make list of scores at drms for this chip set
                    //tw.WriteLine(string.Join("\n", 
                    //                         drmMapTssAllDrms.Select(mapEntry => "list" + mapEntry.Key + "_" + i +" <- list(" + 
                    //                        string.Join(", ",  mapEntry.Value.Where(drm => drmIndex.ContainsKey(drm)).Select(drm => drmScores[i][drmIndex[drm]]))
                    //        + ")")));

                }

                foreach (var gene in v)
                {
                    tw.WriteLine("list" + gene.Tss + " <- list(" + string.Join(",\n", drmScores.Select((x, i) => "list" + gene.Tss + "_" + i)) + ")");
                }
                //tw.WriteLine(string.Join("\n", drmMapTssAllDrms.Select(mapEntry => "list" + mapEntry.Key + " <- list(" + string.Join(", ", drmScores.Select((x, i) => "list" + mapEntry.Key + "_" + i)) + ")")));
                tw.WriteLine("return (list(" + string.Join(",\n", drmMapTssAllDrms.Select(mapEntry => "list" + mapEntry.Key)) + ")) }");
            }*/

            return batchFiles;
        }

        public static int CreateTssScoreTable(
            List<double[]> tssScores, 
            Dictionary<string, int> tssIndex,
            Dictionary<string, double> expression,
            string outFileName, 
            List<string> chipSets)
        {
            var v = expression.Where(tss => tssIndex.ContainsKey(tss.Key))
                .Select(tss => new
                    {
                        Tss = tss.Key,
                        TssOnlyScore = tssScores.Select(tssScore => tssScore[tssIndex[tss.Key]]),
                        Exp = tss.Value
                });

            Console.WriteLine("Number of TSSes: {0}", v.Count());
            
            using (TextWriter tw = new StreamWriter(outFileName))
            {
                // table format
                tw.WriteLine("{0}\tesc", 
                             string.Join("\t", chipSets.Select(x => string.Format("{0}", x.Trim().Split('/').Last().Split('.').First().Split('-')[0]))));
                foreach (var gene in v)
                {
                    tw.WriteLine("{0}\t{1}\t{2}", gene.Tss, string.Join("\t", gene.TssOnlyScore.Select(x => x >= 0 ? x : 0.0)), gene.Exp);
                }
            }
            
            return v.Count();
        }

        /// <summary>
        /// Allowed regression functions
        /// </summary>
        public enum RegressionFunction
        {
            Linear,
            PrincipalComponents,
            PrincipalComponentsCrossValidated,
        };

        /// <summary>
        /// Regresses function R names
        /// </summary>
        private static Dictionary<RegressionFunction, string> regressionFunctions = new Dictionary<RegressionFunction, string>
        {
            { RegressionFunction.Linear, "regress{0}" },
            { RegressionFunction.PrincipalComponents, "regress.pcr{0}" },
            { RegressionFunction.PrincipalComponentsCrossValidated, "regress.pcr.cv{0}" },
        };

        private static string GetFunction(RegressionFunction selection, string suffix)
        {
            return string.Format(regressionFunctions[selection], suffix);
        }

        /// <summary>
        /// Gets hash delimited data from R results
        /// </summary>
        /// <returns>The hash delimited data.</returns>
        /// <param name="filename">Filename.</param>
        /// <param name="function">Function.</param>
        private static double[] GetHashDelimitedData(string filename, RegressionFunction function)
        {
            Func<string[], int, int> featureParse = (fields, i) => function != RegressionFunction.PrincipalComponentsCrossValidated ? 1 : int.Parse(fields[i]);
            
            using (TextReader tr = new StreamReader(filename))
            {
                Regex rSquareds = new Regex("#.+#.+#");
                Regex printCommand = new Regex("sprintf");
                var output = tr.ReadToEnd().Split('\n')
                    //.Select(line => { Console.WriteLine(line); return line; })
                    .Where(line => rSquareds.IsMatch(line) && !printCommand.IsMatch(line))
                        .Select(line => 
                        {
                            string[] vals = rSquareds.Match(line).Value.Split('#').ToArray();
                            //Console.WriteLine(line);
                            return new
                            {
                                Pos = int.Parse(vals[1]),
                                RSquared = double.Parse(vals[2]),
                                NFeatures = featureParse(vals, 3)
                            };
                        })
                        .GroupBy(x => x.Pos)
                        .Select(x => new { Pos = x.Key, RSquared = x.Average(y => y.RSquared) });

                
                double[] data = new double[output.Count()];

                foreach (var point in output)
                {
                    data[point.Pos] = point.RSquared;
                }

                return data;
            }
        }

        /// <summary>
        /// Regresses the chip set columns names against expression (labeled 'esc')
        /// </summary>
        /// <returns>The data.</returns>
        /// <param name="outFileName">Out file name.</param>
        /// <param name="chipSets">Chip sets.</param>
        /// <param name="function">Function.</param>
        public static double[] RegressData(string outFileName, List<string> chipSets, RegressionFunction function)
        {
            string scriptName = "tempReg.R";
            using (TextWriter tw = new StreamWriter(scriptName))
            {
                var columns = chipSets.Select(x => x.Trim().Split('/').Last().Split('.').First().Split('-') [0]);
                
                WriteSource(tw);
                
                tw.WriteLine("columns <- c(\"{0}\");", string.Join("\", \"", columns)); 
                if (function == RegressionFunction.Linear)
                {
                    tw.WriteLine("formula <- as.formula(paste(\"esc ~ \", paste(columns, collapse = \"+\" )))");
                    
                    tw.WriteLine("{0}(\"{1}\", formula)", GetFunction(function, ""), outFileName);
                }
                
                tw.WriteLine("q(save=\"no\")");
            }

            string outputFile = "out.txt";
            RunScript(scriptName, outputFile);
            
            return GetHashDelimitedData(outputFile, function);
        }

        /// <summary>
        /// Gets named regression weights of input features
        /// </summary>
        /// <returns>The data.</returns>
        /// <param name="outFileName">Out file name.</param>
        /// <param name="chipSets">Chip sets.</param>
        /// <param name="function">Function.</param>
        public static Dictionary<string, double> GetRegressWeights(string outFileName, List<string> chipSets)
        {
            string scriptName = "tempReg.R";
            using (TextWriter tw = new StreamWriter(scriptName))
            {
                var columns = chipSets.Select(x => x.Trim().Split('/').Last().Split('.').First().Split('-') [0]);
                
                WriteSource(tw);
                
                tw.WriteLine("columns <- c(\"{0}\");", string.Join("\", \"", columns)); 

                tw.WriteLine("formula <- as.formula(paste(\"esc ~ \", paste(columns, collapse = \"+\" )))");
                 
                tw.WriteLine("file <- \"{0}\"", outFileName);
                tw.WriteLine("x = read.table(file)");
                tw.WriteLine("xn = log(x + 0.01)");
                
                tw.WriteLine("r = lm(formula, data=xn)");
                tw.WriteLine("fixNa <- function(model) { for ( i in 1:length(model$coefficients)) { if (is.na(model$coefficients[i])) { model$coefficients[i] = 0; } }; return (model); };");
                tw.WriteLine("fixedR <- fixNa(r)");
                tw.WriteLine("for (i in 1:length(columns)) { print(sprintf(\"#%s#%f#\", columns[i], fixedR$coefficients[i+1])); }");

                
                tw.WriteLine("q(save=\"no\")");
            }
            
            string outputFile = Path.GetTempFileName();
            RunScript(scriptName, outputFile);
            
            using (TextReader tr = new StreamReader(outputFile))
            {
                Regex coeffs = new Regex("^\\[.*#.+#.+#");

                return tr.ReadToEnd().Split('\n')
                    .Where(line => !string.IsNullOrWhiteSpace(line))
                    .Where(line => coeffs.IsMatch(line))
                    .Select(line => 
                    {
                        var fields = line.Split('#');
                        return new 
                        {
                            DataSet = fields[1],
                            Coeff = double.Parse(fields[2])
                            };
                    }).ToDictionary(x => x.DataSet, x => x.Coeff);
            }
        }


        public static double[] RegressGeneSetOneFeatureWithNNAndSampledControl(string outFileName, List<string> dataSets, List<string> geneSet)
        {
            string scriptName = "temp1freg.R";
            using (TextWriter tw = new StreamWriter(scriptName))
            {
                tw.WriteLine("source(\"" + ScriptPath + "\"); library(boot);");

                tw.WriteLine("geneSet <- c(\"{0}\")", geneSet [0]);
                for (int i = 1; i < geneSet.Count; i++)
                {
                    tw.WriteLine("geneSet <- c(geneSet, \"{0}\")", geneSet[i]);
                }

                tw.WriteLine("columns <- c(\"{0}\");", string.Join("\", \"", dataSets)); 

                tw.WriteLine("TSSformula <- as.formula(paste(\"esc ~ \", columns[1]))");
                tw.WriteLine("MapDRMformula <- as.formula(paste(\"esc ~ \", columns[2]))");
                tw.WriteLine("NNDRMformula <- as.formula(paste(\"esc ~ \", columns[3]))");

                tw.WriteLine("file <- \"{0}\"", outFileName);

                tw.WriteLine("x = read.table(file)");
                tw.WriteLine("xn = log(x + 0.01)");

                tw.WriteLine("fixNa <- function(model) { for ( i in 1:length(model$coefficients)) { if (is.na(model$coefficients[i])) { model$coefficients[i] = 0; } }; return (model); };");

                tw.WriteLine("r_TSS = fixNa(lm(TSSformula, data=xn[geneSet,]))");
                tw.WriteLine("r_MapDRM = fixNa(lm(MapDRMformula, data=xn[geneSet,]))");
                tw.WriteLine("r_NNDRM = fixNa(lm(NNDRMformula, data=xn[geneSet,]))");

                tw.WriteLine("ra_TSS = summary(r_TSS)");
                tw.WriteLine("ra_MapDRM = summary(r_MapDRM)");
                tw.WriteLine("ra_NNDRM = summary(r_NNDRM)");

                tw.WriteLine("y <- as.matrix(xn[geneSet,]$esc); xm <- as.matrix(xn[geneSet,columns[1]]);");
                tw.WriteLine("res <- y - r_TSS$coefficients[1] - xm %*% as.matrix(r_TSS$coefficients[columns[1]])");
                tw.WriteLine("res2 <- res^2");
                tw.WriteLine("tot2 <- (y - mean(y))^2");
                tw.WriteLine("sampleSize = length(y)");
                tw.WriteLine("fstat <- function(x, d) { r2 <- 1 - sum(res2[d]) / sum(tot2[d]); adjr2 <- 1 - (1 - r2) * (length(y) - 1) / (length(y) - length(columns[1]) - 1); return(adjr2);  }");
                tw.WriteLine("repCount <- 1000");
                tw.WriteLine("ssr2 <- boot(y, fstat, repCount)");

                tw.WriteLine("m <- mean(ssr2$t)");

                tw.WriteLine("dv <- sqrt(var(ssr2$t))");

                tw.WriteLine("z <- 0; if (dv != 0) { z <- (ra_MapDRM$adj.r.squared - m) / dv }");
                tw.WriteLine("zNN <- 0; if (dv != 0) { zNN <- (ra_NNDRM$adj.r.squared - m) / dv}");

                tw.WriteLine("bootlm <- function(x, d) { rd <- fixNa(lm(TSSformula, data=xn[d[1:sampleSize],])); rda <- summary(rd); return (rda$adj.r.squared); }");
                tw.WriteLine("SampledTSSRs <- boot(xn[,]$esc, bootlm, repCount)");
                tw.WriteLine("bootlm <- function(x, d) { rd <- fixNa(lm(NNDRMformula, data=xn[d[1:sampleSize],])); rda <- summary(rd); return (rda$adj.r.squared); }");
                tw.WriteLine("SampledNNDRMRs <- boot(xn[,]$esc, bootlm, repCount)");

                tw.WriteLine("ms <- mean(SampledTSSRs$t)");
                tw.WriteLine("dvs <- sqrt(var(SampledTSSRs$t))");
                tw.WriteLine("msnn <- mean(SampledNNDRMRs$t)");
                tw.WriteLine("zMSNN <- 0; if (dvs != 0) { zMSNN <- (msnn - ms) / dvs; }");

                tw.WriteLine("print(sprintf(\"#0#%f#\", z))");
                tw.WriteLine("print(sprintf(\"#1#%f#\", zNN))");
                tw.WriteLine("print(sprintf(\"#2#%f#\", zMSNN))");
                tw.WriteLine("print(sprintf(\"#3#%f#\", ra_TSS$adj.r.squared))");
                tw.WriteLine("print(sprintf(\"#4#%f#\", ra_MapDRM$adj.r.squared))");
                tw.WriteLine("print(sprintf(\"#5#%f#\", ra_NNDRM$adj.r.squared))");
                tw.WriteLine("print(sprintf(\"#6#%f#\", ms))");
                tw.WriteLine("print(sprintf(\"#7#%f#\", msnn))");

                
                tw.WriteLine("q(save=\"no\")");
            }

            string outputFile = "out.txt";
            RunScript(scriptName, outputFile);
            
            return GetHashDelimitedData(outputFile, RegressionFunction.Linear);
        }

        /// <summary>
        /// Regress using all TSS in genome and evaluate on subsets in the map (or of equivalent size randomly sampled)
        /// </summary>
        /// <returns>The data evaluate subsets.</returns>
        /// <param name="outFileName">Out file name.</param>
        /// <param name="chipSets">Chip sets.</param>
        /// <param name="function">Function.</param>
        /// <param name="geneSets">Gene sets.</param>
        /// <param name="Dev">Dev.</param>
        /// <param name="Pval">Pval.</param>
        public static double[] RegressDataEvaluateSubsets(string outFileName, List<string> chipSets, RegressionFunction function, List<List<string>> geneSets, out double[] Mean, out double[] Dev, out double[] Pval)
        {
            string scriptName = "tempReg.R";
            using (TextWriter tw = new StreamWriter(scriptName))
            {
                var columns = chipSets.Select(x => x.Trim().Split('/').Last().Split('.').First().Split('-') [0]);
                
                WriteSource(tw);
                tw.WriteLine("library(boot);");
                
                tw.WriteLine("columns <- c(\"{0}\");", string.Join("\", \"", columns)); 
                if (function == RegressionFunction.Linear)
                {
                    tw.WriteLine("formula <- as.formula(paste(\"esc ~ \", paste(columns, collapse = \"+\" )))");

                    for (int i = 0; i < geneSets.Count; i++)
                    {
                        tw.WriteLine("g{0} <- c(\"{1}\")", i, geneSets[i][0]);
                        for (int j = 1; j < geneSets[i].Count; j++)
                        {
                            tw.WriteLine("g{0} <- c(g{0}, \"{1}\")", i, geneSets[i][j]);
                        }
                    }

                    tw.WriteLine("genes <- list({0})", string.Join(",", geneSets.Select((x, i) => string.Format("g{0} = g{0}", i))));

                    tw.WriteLine("file <- \"{0}\"", outFileName);
                    tw.WriteLine("x = read.table(file)");
                    tw.WriteLine("xn = log(x + 0.01)");

                    tw.WriteLine("r = lm(formula, data=xn)");

                    tw.WriteLine("ra = summary(r)");
                    tw.WriteLine("adjR <- sprintf(\"#Map#0#%f#\", ra[9])");

                    tw.WriteLine("print(adjR)");

                    for (int i = 0; i < geneSets.Count; i++)
                    {
                        tw.WriteLine("y <- as.matrix(xn[genes$g{0},]$esc)", i);
                        tw.WriteLine("xm <- as.matrix(xn[genes$g{0},columns])", i);
                        tw.WriteLine("for (i in 1:length(columns)) { if (is.na(r$coefficients[columns[i]])) { r$coefficients[columns[i]] = 0; } }");
                        tw.WriteLine("res <- y - r$coefficients[1] - xm %*% as.matrix(r$coefficients[columns])");
                        tw.WriteLine("res2 <- apply(res^2, 2, sum)");

                        tw.WriteLine("SStot <- sum((y - mean(y))^2)");

                        tw.WriteLine("NARsq <- 1 - res2 / SStot");

                        tw.WriteLine("yall <- as.matrix(xn[,]$esc); xmall <- as.matrix(xn[,columns]);");

                        tw.WriteLine("SSadjr2 <- 1 - (1 - NARsq) * (length(yall) - 1) / (length(yall) - length(columns) - 1)");
                        tw.WriteLine("Rsq <- sprintf(\"#Map#{0}#%f#\", SSadjr2)", i + 1);

                        tw.WriteLine("print(Rsq)");

                        tw.WriteLine("sampleSize = length(y)");

                        tw.WriteLine("yall <- as.matrix(xn[,]$esc); xmall <- as.matrix(xn[,columns]);");
                        tw.WriteLine("resall <- yall - r$coefficients[1] - xmall %*% as.matrix(r$coefficients[columns])");
                        tw.WriteLine("res2all <- resall^2");
                        tw.WriteLine("tot2all <- (yall - mean(yall))^2");
                        tw.WriteLine("fstat <- function(x, d) { samp = sample.int(length(res2all))[1:sampleSize]; r2 <- 1 - sum(res2all[samp]) / sum(tot2all[samp]); adjr2 <- 1 - (1 - r2) * (length(yall) - 1) / (length(yall) - length(columns) - 1); return(adjr2);  }");
                        tw.WriteLine("repCount <- 10000");
                        // Use Y as the boot data arg to get the right data set size to sample
                        tw.WriteLine("ssr2 <- boot(y, fstat, repCount)");
                        tw.WriteLine("countlt <- 0;");
                        tw.WriteLine("for (k in 1:repCount) { if (ssr2$t[k] < SSadjr2) { countlt <- countlt + 1; } }");
                        tw.WriteLine("dv <- sqrt(var(ssr2$t))");
                        tw.WriteLine("mnn <- mean(ssr2$t)");
                        tw.WriteLine("print(sprintf(\"#Mean#{0}#%f#\", mnn))", i + 1);
                        tw.WriteLine("print(sprintf(\"#Dev#{0}#%f#\", dv))", i + 1);
                        tw.WriteLine("print(sprintf(\"#Pval#{0}#%f#\", countlt / repCount))", i + 1);
                    }
                }

                tw.WriteLine("q(save=\"no\")");
            }
            
            string outputFile = "out.txt";
            RunScript(scriptName, outputFile);

            double[] Rsquareds;
            using (TextReader tr = new StreamReader(outputFile))
            {
                Regex rSquareds = new Regex("^\\[.*#Map#.+#.+#");
                Regex rSquaredMean = new Regex("^\\[.*#Mean#.+#.+#");
                Regex rSquaredDev = new Regex("^\\[.*#Dev#.+#.+#");
                Regex pvals = new Regex("^\\[.*#Pval#.+#.+#");

                var lines = tr.ReadToEnd().Split('\n');

                Func<Regex, double[]> GetValues = regex => 
                {
                    var Rdata = lines.Where(line => regex.IsMatch(line)).Select(line => 
                    {
                        var fields = line.Split('#');
                        return new 
                        {
                            Index = int.Parse(fields[2]),
                            Value = double.Parse(fields[3])
                        };
                    });
                
                    double[] data = new double[geneSets.Count + 1];

                    foreach (var elem in Rdata)
                    {
                        data[elem.Index] = elem.Value;
                    }

                    return data;
                };

                Rsquareds = GetValues(rSquareds);
                Mean = GetValues(rSquaredMean);
                Dev = GetValues(rSquaredDev);
                Pval = GetValues(pvals);
            }

            return Rsquareds;
        }

        private static void MakeColumns(TextWriter tw, List<string> chipSets)
        {
            var columns = chipSets.Select(x => x.Trim().Split('/').Last().Split('.').First().Split('-')[0]);

            tw.WriteLine("columnsTSS <- c(\"{0}\");", string.Join("\", \"", columns.Select(x => string.Format("Tss_{0}", x)))); 
            tw.WriteLine("columnsDRM <- c(\"{0}\");", string.Join("\", \"", columns.Select(x => string.Format("Drm_{0}", x)))); 
            tw.WriteLine("columns <- c(columnsTSS, columnsDRM);");
        }

        /// <summary>
        /// TSS column header names
        /// </summary>
        /// <returns>The tss column names.</returns>
        /// <param name="chipDataSets">Chip data sets.</param>
        public static IEnumerable<string> MakeTssColumnNames(List<string> chipDataSets)
        {
            return chipDataSets.Select(x => string.Format("Tss_{0}", x));
        }

        /// <summary>
        /// DRM column header names
        /// </summary>
        /// <returns>The tss column names.</returns>
        /// <param name="chipDataSets">Chip data sets.</param>
        public static IEnumerable<string> MakeDrmColumnNames(List<string> chipDataSets)
        {
            return chipDataSets.Select(x => string.Format("Drm_{0}", x));
        }

        /// <summary>
        /// Row count parameter for regression on all rows
        /// </summary>
        /// <value>All rows.</value>
        public class RegressColumnsParam { public static int AllRows { get { return -1; } } }

        /// <summary>
        /// Create an output file name for a batch run of a script
        /// </summary>
        /// <returns>The script batch output file.</returns>
        /// <param name="scriptName">Script name.</param>
        /// <param name="batch">Batch.</param>
        public static string MakeScriptBatchOutputFile(string scriptName, int batch)
        {
            return scriptName.Replace(".R", "." + batch + ".Rout");
        }

        /// <summary>
        /// Regress a set of columns versus expression in a given table data file
        /// </summary>
        /// <returns>The columns.</returns>
        /// <param name="scriptName">script file to generate to run regression</param>
        /// <param name="tableFile">Data file name of regression input data</param>
        /// <param name="columns">Column names to regress</param>
        /// <param name="rowCount">Count of rows to sample; -1 for all </param> 
        /// <param name="repCount">number of R2 samples</param>
        /// <param name="batch">batch number of R2 samples to generate</param>
        public static double[] RegressColumns(
            string scriptName, 
            string tableFile,
            IEnumerable<string> columns, 
            int rowCount, 
            int repCount,  
            int batch)
        {
            using (TextWriter tw = new StreamWriter(scriptName))
            {
                WriteSource(tw);
                
                tw.WriteLine("columns <- c(\"{0}\");", string.Join("\", \"", columns));
                tw.WriteLine("file <- \"" + tableFile + "\"");
                tw.WriteLine("repCount <- " + repCount);

                tw.WriteLine("r2filename <- \"" + scriptName.Replace(".R", "." + batch + ".R2s") + "\"");
                if (rowCount == RegressColumnsParam.AllRows)
                {
                    tw.WriteLine("regress.columns.r2sample(file, columns, repCount, r2filename)");
                }
                else
                {
                    tw.WriteLine("rowCount <- " + rowCount);
                    tw.WriteLine("regress.columns.rowsample(file, columns, repCount, rowCount, r2filename)");
                }

                tw.WriteLine("q(save=\"no\")");
            }

            string outputFile = MakeScriptBatchOutputFile(scriptName, batch);
            RunScript(scriptName, outputFile);

            return ReadR2s(scriptName, new string[] { outputFile } );
        }

        /// <summary>
        /// Get a single batch's R2 values
        /// </summary>
        /// <returns>The r2 set.</returns>
        /// <param name="scriptName">Script name.</param>
        /// <param name="batch">Batch.</param>
        public static double[] ReadR2Set(
            string scriptName,
            int batch)
        {
            return ReadR2s(scriptName, new string[] { MakeScriptBatchOutputFile(scriptName, batch) });
        }

        /// <summary>
        /// Read R2 data from all or a selected batch
        /// </summary>
        /// <returns>The r2s.</returns>
        /// <param name="scriptName">Script name.</param>
        /// <param name="batchCount">Batch count.</param>
        public static double[] ReadR2s(
            string scriptName, 
            string[] outputFiles)
        {
            // | Set Index | R-squared | 
            Regex setRs = new Regex("Rsquareds\\|.*\\|.*\\|");

            //Func<int, string> getBatchFilename = i => scriptName.Replace(".R", "." + i + ".Rout");
            Func<TextReader, IEnumerable<double>> parseFile = tr => 
            {
                return tr.ReadToEnd().Split('\n')
                    .Where(line => setRs.IsMatch(line))
                    .Select(line => double.Parse(line.Trim().Split('|')[2]));
            };
        
            return outputFiles
                .Select(x => 
                {
                    using (TextReader tr = new StreamReader(x))
                    {
                        return parseFile(tr);
                    }
                })
                .SelectMany(x => x).ToArray();
        }

        /// <summary>
        /// Regress promoter only model and shuffled DRM scores
        /// </summary>
        /// <returns>The prom and shuffled drm.</returns>
        /// <param name="dataFileName">Data file name.</param>
        /// <param name="chipDataSets">Chip data sets.</param>
        /// <param name="pvalue">Pvalue.</param>
        public static double[] RegressPromAndShuffledDrm(string dataFileName, List<string> chipDataSets, out double pvalue)
        {
            string scriptName = "../temp/scriptShuffleDrm.R";
            using (TextWriter tw = new StreamWriter(scriptName))
            {
                WriteSource(tw);
                
                MakeColumns(tw, chipDataSets);
                
                tw.WriteLine("regress.promandmap(\"" + dataFileName + "\", columnsTSS, columnsDRM, 10)");
                
                tw.WriteLine("q(save=\"no\")");
            }
            
            string outputFile = "../temp/promout.txt"; //Path.GetTempFileName();
            RunScript(scriptName, outputFile);
            
            using (TextReader tr = new StreamReader(outputFile))
            {
                // | Set Index | R-squared | 
                Regex setRs = new Regex("Rsquareds\\|.*\\|.*\\|");

                // | pvalue |
                Regex setSetPvalues = new Regex("pvalue\\|.*\\|");

                var lines = tr.ReadToEnd().Split('\n').Where(line => setRs.IsMatch(line) || setSetPvalues.IsMatch(line)).Select(x => x.Trim());

                pvalue = lines.Where(line => setSetPvalues.IsMatch(line)).Select(line => 
                                                                        {
                    var fields = line.Split('|');

                    return double.Parse(fields[1]);
                }).First();

                return lines.Where(line => setRs.IsMatch(line)).Select(line =>
                                                                       {
                        var fields = line.Split('|');
                        
                        return new 
                        {
                            Index = int.Parse(fields [1]),
                            Rsquared = double.Parse(fields [2])
                        };
                    })
                    .OrderBy(x => x.Index)
                    .Select(x => x.Rsquared)
                    .ToArray();
            }
        }

        /// <summary>
        /// Regresses the data using the specified function
        /// </summary>
        /// <returns>R-squared values of the regression of different types</returns>
        /// <param name="outFileName">Out file name.</param>
        /// <param name="chipSets">Chip sets.</param>
        /// <param name="function">Function.</param>
        public static double[] RegressMapScoreData(string outFileName, List<string> chipSets, RegressionFunction function)
        {
            string scriptName = "tempReg.R";
            using (TextWriter tw = new StreamWriter(scriptName))
            {
                WriteSource(tw);

                MakeColumns(tw, chipSets);
               
                if (function == RegressionFunction.Linear)
                {
                    tw.WriteLine("formulaTSS <- as.formula(paste(\"esc ~ \", paste(columnsTSS, collapse = \"+\" )))");
                    tw.WriteLine("formulaDRM <- as.formula(paste(\"esc ~ \", paste(columnsDRM, collapse = \"+\" )))");

                    tw.WriteLine("formula <- as.formula(paste(\"esc ~ \", paste(columns\t, collapse = \"+\" )))");

                    tw.WriteLine("    {0}(\"{1}\", formula, formulaTSS, formulaDRM, {2}, {3})", GetFunction(function, ".drm"), outFileName);
                }
                else 
                {
                    tw.WriteLine("{0}(\"{1}\", columns, columnsTSS, columnsDRM)", GetFunction(function, ".drm"), outFileName);
                }
                
                tw.WriteLine("q(save=\"no\")");
            }

            string outputFile = "out.txt";
            RunScript(scriptName, outputFile);

            return GetHashDelimitedData(outputFile, function);
        }
       
        /// <summary>
        /// Regress on TSS-DRM maps and controls getting p-values of difference in R-squared through sub-sampling residuals or gene sets
        /// </summary>
        /// <param name="outFileNames">Out file names.</param>
        /// <param name="chipDataSets">Chip data sets.</param>
        /// <param name="repCount">Rep count.</param>
        /// <param name="Rsquareds">Rsquareds.</param>
        /// <param name="pvals">Pvals.</param>
        public static void BootRegressMapScoreData(
            string[] outFileNames, 
            IEnumerable<string> listSampleFiles, 
            List<string> chipDataSets, 
            int repCount, 
            string baseSciptFilename,
            out double[] Rsquareds, 
            out double[][] pvals)
        {
            string scriptName = baseSciptFilename;
            using (TextWriter tw = new StreamWriter(scriptName))
            {
                WriteSource(tw);
                
                MakeColumns(tw, chipDataSets);

                tw.WriteLine("fileMap <- \"" + outFileNames[0] + "\"");
                tw.WriteLine("fileNN <- \"" + outFileNames[1] + "\"");
                tw.WriteLine("fileNNR <- \"" + outFileNames[2] + "\"");
                tw.WriteLine("fileDR <- \"" + outFileNames[3] + "\"");
                tw.WriteLine("listSampleFiles <- list("+ (listSampleFiles.Count() > 0 ? ("\"" +  string.Join("\", \"", listSampleFiles) + "\"") : "") +")");
                tw.WriteLine("regress.drm.boot(fileMap, fileNN, fileNNR, fileDR, listSampleFiles, columnsTSS, columnsDRM, " + repCount + ", " + batchSize + ")");


                tw.WriteLine("q(save=\"no\")");
            }

            string outputFile = baseSciptFilename.Replace(".R", ".Rout");
            RunScript(scriptName, outputFile);

            using (TextReader tr = new StreamReader(outputFile))
            {
                // | Set Index | Subset Index | R-squared | 
                Regex setRs = new Regex("Rsquareds\\|.*\\|.*\\|.*\\|");

                // | Set1 Index | Subset1 Index | Set2 Index | Subset2 Index | pvalue |
                Regex setSetPvalues = new Regex("pvalue\\|.*\\|.*\\|.*\\|.*\\|.*\\|");

                var lines = tr.ReadToEnd().Split('\n').Where(line => setRs.IsMatch(line) || setSetPvalues.IsMatch(line)).Select(x => x.Trim());

                Func<string, string, int> MakeIndex = (f1, f2) => ((int.Parse(f1) - 1) * 3 + (int.Parse(f2) - 1));

                Rsquareds = lines.Where(line => setRs.IsMatch(line)).Select(line =>
                {
                    var fields = line.Split('|');

                    return new 
                    {
                        Index = MakeIndex(fields[1], fields[2]),
                        Rsquared = double.Parse(fields[3])
                    };
                })
                .OrderBy(x => x.Index)
                .Select(x => x.Rsquared)
                .ToArray();

                var pvalData = lines.Where(line => setSetPvalues.IsMatch(line)).Select(line => 
                {
                    var fields = line.Split('|');

                    return new 
                    {
                        Index1 = MakeIndex(fields[1], fields[2]),
                        Index2 = MakeIndex(fields[3], fields[4]),
                        Pvalue = double.Parse(fields[5])
                    };
                })
                .ToLookup(x => x.Index1, x => x)
                .ToDictionary(x => x.Key, x => x.ToDictionary(y => y.Index2, y => y.Pvalue));

                int dim = 10;

                pvals = new double[dim][];

                for (int i = 0; i < dim; i++)
                {
                    pvals[i] = new double[dim];
                    for (int j = 0; j < dim; j++)
                    {
                        if (pvalData.ContainsKey(i) && pvalData[i].ContainsKey(j))
                        {
                            pvals[i][j] = pvalData[i][j];
                        }
                        else
                        {
                            pvals[i][j] = double.NaN;
                        }
                    }
                }
            }
        }

        private static Regex extractPval = new Regex("\\[1\\] \"#.*#.*#");

        public static double[][] TestCorrelations(IEnumerable<Tuple<string, string>> vectorcsv)
        {
            string scriptName = "tempCorTest.R";
            using (TextWriter tw = new StreamWriter(scriptName))
            {
                foreach (var vector in vectorcsv)
                {
                    tw.WriteLine("x <- c({0})", vector.Item1);
                    tw.WriteLine("y <- c({0})", vector.Item2);
                    tw.WriteLine("line <- sprintf(\"#%e#%f#\", cor.test(x, y)$p.value, cor.test(x, y)$statistic)");
                    tw.WriteLine("print(line)");
                }

                tw.WriteLine("q(save=\"no\")");
            }

            string outputFile = "ctout.txt";
            RunScript(scriptName, outputFile);

            using (TextReader tr = new StreamReader(outputFile))
            {
                return tr.ReadToEnd().Split('\n')
                    .Select(line => extractPval.Matches(line))
                    .Where(m => m.Count > 0)
                    .Select(m =>
                    {
                        var a = m[0].Value.Split('#');
                        //Console.WriteLine(m[0]);
                        return new double[] { double.Parse(a[1]), double.Parse(a[2]) };
                    })
                    .ToArray();
            }
        }

        private static void RunScript(string scriptFile, string outputFile)
        {
            System.Diagnostics.ProcessStartInfo info = new System.Diagnostics.ProcessStartInfo();
            info.UseShellExecute = false;
            info.CreateNoWindow = true;
            
            info.RedirectStandardOutput = true;
            info.WorkingDirectory = ".";
            info.FileName = "R";
            info.Arguments = string.Format("CMD BATCH {0} {1}", scriptFile, outputFile);
            
            System.Diagnostics.Process proc = System.Diagnostics.Process.Start(info);
            
            proc.WaitForExit();

            // Clear out .RData 
            /*info.FileName = "rm";
            info.Arguments = ".RData";
            proc = System.Diagnostics.Process.Start(info);
            
            proc.WaitForExit();*/
        }

        public static void PlotHistoneProfile(string filename, string cellLine, string expressionType, string histoneType)
        {
            string createPng = 
                "png(file=\"results/HistoneProfile/profiles/" + cellLine + "." + expressionType + "." + histoneType + ".png\", width=600, height=400)\n";
            
            string createPdf = 
                "pdf(file=\"results/HistoneProfile/profiles/" + cellLine + "." + expressionType + "." + histoneType + ".pdf\", width=7.5, height=5)\n";
            
            string load = 
                "x <- read.table(\"" + filename + "\")\n" +
                "ccc <- c(rgb(0,1,0,0.8), rgb(0,1,1,0.8))\n";

            string plot = 
                "ymax <- 0.75; if (max(x[,2:3]) > ymax) { ymax <- 0.75 }\n" +
                "plot(x[,1], x[,2], col=ccc[1], ylim=c(0, ymax), xlim=c(-10000, 10000), xlab=\"Distance from TSS\", ylab=\"Fraction of Peaks\")\n" +
                "points(x[,1], x[,3], col=ccc[2])\n" +
                "title(\"Promoter Distribution of " + histoneType + " peaks in " + cellLine + " " + expressionType + " Map\")\n" +
                "legend(\"topleft\", title=\"TSS Set\", c(\"Variantly Expressed\", \"Map\"), fill=ccc)\n" +
                "dev.off()\n";

            string scriptName = "../temp/plotHistoneProfile" + cellLine + expressionType + histoneType + ".R";

            using (TextWriter tw = new StreamWriter(scriptName))
            {
                tw.WriteLine(load + createPng + plot + createPdf + plot);

                
                tw.WriteLine("q(save=\"no\")");
            }

            RunScript(scriptName, "../temp/junk");

        }

        /// <summary>
        /// Writes the source invocation for the R source
        /// </summary>
        /// <param name="tw">Tw.</param>
        private static void WriteSource(TextWriter tw)
        {
            tw.WriteLine("source(\"" + ScriptPath + "\");");
        }
    }
}

